//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLKernel.hpp"
#include <sys/time.h>
#include <unistd.h>
#include <algorithm>

static long long getTimestamp(void)
{
	static __thread struct timeval tv;
	gettimeofday(&tv, NULL);
	return 1000000LL * tv.tv_sec + tv.tv_usec;
}

ArmCLKernel::ArmCLKernel(cl_kernel kernel, const char* kernel_name, ArmCLProgram* armcl_program)
    : kernel_(kernel), kernel_name_(kernel_name), armcl_program_(armcl_program)
{
    program_ = armcl_program_->program();
    armcl_context_ = armcl_program->armcl_context();
    context_ = armcl_context_->context();

    PrepareOpenCLCalls();
    PrepareContextInformation();

    PrepareThrottlingKernels();
}

ArmCLKernel::~ArmCLKernel()
{
}

cl_int ArmCLKernel::SetKernelArgSVMPointer(cl_uint arg_index, const void* arg_value)
{
    cl_int err;

    if (onthefly_) {
        if (onthefly_kernel_cpu_thread_) {
            for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                int index = i * onthefly_kernel_cpu_thread_num_args_ + (int)arg_index;
                onthefly_kernel_cpu_thread_params_[index] = (unsigned long)arg_value;
            }
        }

        if (onthefly_kernel_cpu_thread_array_[0]) {
            for (int j = 0; j < 100; j++) {
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * onthefly_kernel_cpu_thread_num_args_ + (int)arg_index;
                    onthefly_kernel_cpu_thread_array_params_[j][index] = (unsigned long)arg_value;
                }
            }
        }
    }
}

cl_int ArmCLKernel::SetKernelArg(cl_uint arg_index, size_t arg_size,
        const void* arg_value, std::map<cl_mem, ArmCLBuffer*> armcl_buffers)
{
    cl_int err;

    ArmCLBuffer* armcl_buffer = NULL;
    cl_mem* addr_ = (cl_mem*)arg_value;
    cl_mem_flags flags;
    size_t size;
    void* host_ptr = 0;

    for (std::map<cl_mem, ArmCLBuffer*>::iterator it =
            armcl_buffers.begin();
            it != armcl_buffers.end();
            it++) {
        cl_mem memobj = it->first;
        if (memobj == *addr_) {
            armcl_buffer = it->second;

            flags = armcl_buffer->flags();
            size = armcl_buffer->size();
            host_ptr = armcl_buffer->host_ptr();

            break;
        }
    }

    if (throttling_) {
        if (throttling_kernel_) {
            err = clSetKernelArg_p_(throttling_kernel_, arg_index, arg_size, arg_value);
            if (err != CL_SUCCESS)
                std::cout << "ArmCL: clSetKernelArg failed for throttling_kernel" << std::endl;
        }

        if (throttling_kernel_cpu_thread_) {
            if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
                // this argument is opencl buffer (cl_mem)
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * throttling_kernel_cpu_thread_num_args_ + (int)arg_index;
                    throttling_kernel_cpu_thread_params_[index] = (unsigned long)host_ptr;
                }
            } else if (armcl_buffer == NULL) {
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * throttling_kernel_cpu_thread_num_args_ + (int)arg_index;
                    throttling_kernel_cpu_thread_params_[index] = (unsigned long)arg_value;
                }
            }
        }

        if (dynthrottling_kernel_cpu_thread_) {
            if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
                // this argument is opencl buffer (cl_mem)
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * dynthrottling_kernel_cpu_thread_num_args_ + (int)arg_index;
                    dynthrottling_kernel_cpu_thread_params_[index] = (unsigned long)host_ptr;
                }
            } else if (armcl_buffer == NULL) {
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * dynthrottling_kernel_cpu_thread_num_args_ + (int)arg_index;
                    dynthrottling_kernel_cpu_thread_params_[index] = (unsigned long)arg_value;
                }
            }
        }
    }

    if (onthefly_) {
        if (onthefly_kernel_) {
            err = clSetKernelArg_p_(onthefly_kernel_, arg_index, arg_size, arg_value);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: clSetKernelArg failed for onthefly_kernel" << std::endl;
            }
        }

        if (onthefly_kernel_cpu_thread_) {
            if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
                // this argument is opencl buffer (cl_mem)
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * onthefly_kernel_cpu_thread_num_args_ + (int)arg_index;
                    onthefly_kernel_cpu_thread_params_[index] = (unsigned long)host_ptr;
                }
            } else if (armcl_buffer == NULL) {
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * onthefly_kernel_cpu_thread_num_args_ + (int)arg_index;
                    onthefly_kernel_cpu_thread_params_[index] = (unsigned long)arg_value;
                }
            }
        }

        if (onthefly_kernel_cpu_profiler_) {
            if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
                // this argument is opencl buffer (cl_mem)
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * onthefly_kernel_cpu_profiler_num_args_ + (int)arg_index;
                    onthefly_kernel_cpu_profiler_params_[index] = (unsigned long)host_ptr;
                }
            } else if (armcl_buffer == NULL) {
                for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                    int index = i * onthefly_kernel_cpu_profiler_num_args_ + (int)arg_index;
                    onthefly_kernel_cpu_profiler_params_[index] = (unsigned long)arg_value;
                }
            }
        }

        if (onthefly_kernel_cpu_thread_array_[0]) {
            for (int j = 0; j < 100; j++) {
                if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
                    // this argument is opencl buffer (cl_mem)
                    for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                        int index = i * onthefly_kernel_cpu_thread_num_args_ + (int)arg_index;
                        onthefly_kernel_cpu_thread_array_params_[j][index] = (unsigned long)host_ptr;
                    }
                } else if (armcl_buffer == NULL) {
                    for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                        int index = i * onthefly_kernel_cpu_thread_num_args_ + (int)arg_index;
                        onthefly_kernel_cpu_thread_array_params_[j][index] = (unsigned long)arg_value;
                    }
                }
            }
        }
    }

    return 0;
}

void ArmCLKernel::PrepareOpenCLCalls()
{
    char* error;

    clCreateKernel_p_ = (cl_kernel (*) (cl_program program,
                const char* kernel_name,
                cl_int* errcode_ret))dlsym(RTLD_NEXT, "clCreateKernel");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueNDRangeKernel_p_ = (cl_int (*) (cl_command_queue command_queue,
        cl_kernel kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event))dlsym(RTLD_NEXT, "clEnqueueNDRangeKernel");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clSetKernelArg_p_ = (cl_int (*) (cl_kernel kernel,
                cl_uint arg_index,
                size_t arg_size,
                const void* arg_value))
        dlsym(RTLD_NEXT, "clSetKernelArg");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clCreateBuffer_p_ = (cl_mem (*) (cl_context context,
                cl_mem_flags flags,
                size_t size,
                void* host_ptr,
                cl_int* errcode_ret))
        dlsym(RTLD_NEXT, "clCreateBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueWriteBuffer_p_ = (cl_int CL_API_CALL (*) (
                cl_command_queue command_queue,
                cl_mem buffer,
                cl_bool blocking_write,
                size_t offset,
                size_t size,
                const void* ptr,
                cl_uint num_events_in_wait_list,
                const cl_event* event_wait_list,
                cl_event* event))dlsym(RTLD_NEXT, "clEnqueueWriteBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }
}

void ArmCLKernel::PrepareContextInformation()
{
    cl_uint err = clGetProgramInfo(program_,
            CL_PROGRAM_CONTEXT,
            sizeof(cl_context),
            &context_,
            NULL);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: " << "clGetProgramInfo: failed" << std::endl;
    }
}

void ArmCLKernel::PrepareThrottlingKernels()
{
    std::string throttling_kernel_name;
    throttling_kernel_name.append(kernel_name_);
    throttling_kernel_name.append("_throttling");

    cl_int err;
    throttling_kernel_ = clCreateKernel_p_(program_, throttling_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
        std::cerr << "ArmCL: prepare throttling_kernel: can't create kernel "
            << throttling_kernel_name << std::endl;
    }

    clGetKernelInfo(kernel_, CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &kernel_num_args_, NULL);
    clGetKernelInfo(throttling_kernel_, CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &throttling_kernel_num_args_, NULL);

    {
        std::string CPU_thread_kernel_file_name;
        CPU_thread_kernel_file_name.append(kernel_name_);
        CPU_thread_kernel_file_name.append("-throttling.cpp");

        std::string CPU_thread_kernel_library_name;
        CPU_thread_kernel_library_name.append(kernel_name_);
        CPU_thread_kernel_library_name.append("-throttling.so");

        std::string command;
        command.append("g++ -O0 -std=c++11 "); // use c++11 for std::atomic
        command.append(armcl_program_->build_option());
        command.append(" -shared -fpic -o ");
        command.append(CPU_thread_kernel_library_name);
        command.append(" ");
        command.append(CPU_thread_kernel_file_name);

        //std::cout << command << std::endl;
        int ret = system(command.c_str()); // generate dynamic library of CPU kernel thread

        //const char* current_path = std::getenv("PWD");
		char current_path[256];
		getcwd(current_path, 256);
        std::string CPU_thread_kernel_path;
        CPU_thread_kernel_path.append(current_path);
        CPU_thread_kernel_path.append("/");
        CPU_thread_kernel_path.append(CPU_thread_kernel_library_name);
        void* executable = dlopen(CPU_thread_kernel_path.c_str(), RTLD_NOW);
        if (executable == NULL) {
            fprintf(stderr, "executable is NULL\n");
        }

        *(void**)(&throttling_kernel_cpu_thread_) = dlsym(executable, "CPU_thread_entry");
        if (throttling_kernel_cpu_thread_ == NULL) {
            fprintf(stderr, "throttling_kernel_cpu_thread_ is NULL\n");
        }

        throttling_kernel_cpu_thread_num_args_ = (int)kernel_num_args_ + 4;
        throttling_kernel_cpu_thread_params_ = (unsigned long*)dlsym(executable, "x86_kernel_params");
    }

    {
        std::string CPU_thread_kernel_file_name;
        CPU_thread_kernel_file_name.append(kernel_name_);
        CPU_thread_kernel_file_name.append("-dynthrottling.cpp");

        std::string CPU_thread_kernel_library_name;
        CPU_thread_kernel_library_name.append(kernel_name_);
        CPU_thread_kernel_library_name.append("-dynthrottling.so");

        std::string command;
        command.append("g++ -O0 -std=c++11 "); // use c++11 for std::atomic
        command.append(armcl_program_->build_option());
        command.append(" -shared -fpic -o ");
        command.append(CPU_thread_kernel_library_name);
        command.append(" ");
        command.append(CPU_thread_kernel_file_name);

        //std::cout << command << std::endl;
        int ret = system(command.c_str()); // generate dynamic library of CPU kernel thread

        //const char* current_path = std::getenv("PWD");
		char current_path[256];
		getcwd(current_path, 256);
        std::string CPU_thread_kernel_path;
        CPU_thread_kernel_path.append(current_path);
        CPU_thread_kernel_path.append("/");
        CPU_thread_kernel_path.append(CPU_thread_kernel_library_name);
        void* executable = dlopen(CPU_thread_kernel_path.c_str(), RTLD_NOW);
        if (executable == NULL) {
            fprintf(stderr, "executable is NULL\n");
        }

        *(void**)(&dynthrottling_kernel_cpu_thread_) = dlsym(executable, "CPU_thread_entry");
        if (dynthrottling_kernel_cpu_thread_ == NULL) {
            fprintf(stderr, "dynthrottling_kernel_cpu_thread_ is NULL\n");
        }

        dynthrottling_kernel_cpu_thread_num_args_ = (int)kernel_num_args_ + 4;
        dynthrottling_kernel_cpu_thread_params_ = (unsigned long*)dlsym(executable, "x86_kernel_params");
    }
}

void ArmCLKernel::PrepareOnTheFlyKernels()
{
    std::string onthefly_kernel_name;
    onthefly_kernel_name.append(kernel_name_);
    onthefly_kernel_name.append("_OnTheFly");

    std::cout << "ArmCL: " << " prepare profiler run: "
        << " kernel_name: " << kernel_name_
        << " [OnTheFly] onthefly_kernel_name: " << onthefly_kernel_name
        << std::endl;

    cl_int err;
    onthefly_kernel_ = clCreateKernel_p_(program_, onthefly_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: cannot create onthefly kernel " << onthefly_kernel_name << std::endl;
    }
    clGetKernelInfo(kernel_, CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &kernel_num_args_, NULL);
    clGetKernelInfo(onthefly_kernel_, CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &onthefly_kernel_num_args_, NULL);
    onthefly_kernel_cpu_thread_num_args_ = (int)kernel_num_args_ + 5;
    onthefly_kernel_cpu_profiler_num_args_ = (int)kernel_num_args_ + 5;

    std::string CPU_thread_kernel_file_name;
    CPU_thread_kernel_file_name.append(kernel_name_);
    CPU_thread_kernel_file_name.append("-CPU-hand.cpp");

    std::string CPU_thread_kernel_library_name;
    CPU_thread_kernel_library_name.append(kernel_name_);
    CPU_thread_kernel_library_name.append("-CPU-hand.so");

    std::string command;
    command.append("g++ -O0 -std=c++11 -shared -fpic -o "); // use c++11 for std::atomic
    command.append(CPU_thread_kernel_library_name);
    command.append(" ");
    command.append(CPU_thread_kernel_file_name);
    int ret = system(command.c_str()); // generate dynamic library of CPU kernel thread

    //const char* current_path = std::getenv("PWD");
	char current_path[256];
	getcwd(current_path, 256);
    std::string CPU_thread_kernel_path;
    CPU_thread_kernel_path.append(current_path);
    CPU_thread_kernel_path.append("/");
    CPU_thread_kernel_path.append(CPU_thread_kernel_library_name);
    void* executable = dlopen(CPU_thread_kernel_path.c_str(), RTLD_NOW);
    if (executable == NULL) {
        fprintf(stderr, "executable is NULL\n");
    }

    *(void**)(&onthefly_kernel_cpu_thread_) = dlsym(executable, "CPU_thread_entry");
    if (onthefly_kernel_cpu_thread_ == NULL) {
        fprintf(stderr, "onthefly_kernel_cpu_thread_ is NULL\n");
    }

    onthefly_kernel_cpu_thread_params_ = (unsigned long*)dlsym(executable, "x86_kernel_params");
    if (onthefly_kernel_cpu_thread_params_) {
        for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
            for (int j = 0; j < onthefly_kernel_cpu_thread_num_args_; j++) {
                fprintf(stdout, "%lu ", onthefly_kernel_cpu_thread_params_[
                        i*onthefly_kernel_cpu_thread_num_args_ + j]);
            }
            fprintf(stdout, "\n");
        }
    }

    std::string CPU_profiler_kernel_file_name;
    CPU_profiler_kernel_file_name.append(kernel_name_);
    CPU_profiler_kernel_file_name.append("-CPU-profiler-hand.cpp");

    std::string CPU_profiler_kernel_library_name;
    CPU_profiler_kernel_library_name.append(kernel_name_);
    CPU_profiler_kernel_library_name.append("-CPU-profiler-hand.so");

    std::string command_;
    command_.append("g++ -O0 -std=c++11 -shared -fpic -o ");
    command_.append(CPU_profiler_kernel_library_name);
    command_.append(" ");
    command_.append(CPU_profiler_kernel_file_name);
    ret = system(command_.c_str());

    //const char* current_path_ = std::getenv("PWD");
	char current_path_[256];
	getcwd(current_path_, 256);
    std::string CPU_profiler_kernel_path;
    CPU_profiler_kernel_path.append(current_path_);
    CPU_profiler_kernel_path.append("/");
    CPU_profiler_kernel_path.append(CPU_profiler_kernel_library_name);
    void* executable_ = dlopen(CPU_profiler_kernel_path.c_str(), RTLD_NOW);
    if (executable_ == NULL) {
        fprintf(stderr, "executable is NULL\n");
    }

    *(void**)(&onthefly_kernel_cpu_profiler_) = dlsym(executable_, "CPU_thread_entry");
    if (onthefly_kernel_cpu_profiler_ == NULL) {
        fprintf(stderr, "onthefly_kernel_cpu_profiler_ is NULL\n");
    }

    onthefly_kernel_cpu_profiler_params_ = (unsigned long*)dlsym(executable_, "x86_kernel_params");
    if (onthefly_kernel_cpu_profiler_params_) {
        for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
            for (int j = 0; j < onthefly_kernel_cpu_profiler_num_args_; j++) {
                fprintf(stdout, "%lu ", onthefly_kernel_cpu_profiler_params_[
                        i*onthefly_kernel_cpu_profiler_num_args_ + j]);
            }
            fprintf(stdout, "\n");
        }
    }

    onthefly_kernel_cpu_thread_array_params_ = (unsigned long**)malloc(100 * sizeof(unsigned long*));
    for (int i = 0; i <100; i++) {
        std::string CPU_thread_kernel_file_name;
        CPU_thread_kernel_file_name.append(kernel_name_);
        CPU_thread_kernel_file_name.append("-CPU-hand.cpp");

        std::string CPU_thread_kernel_library_name;
        char buf[3];
        sprintf(buf, "%d", i);
        CPU_thread_kernel_library_name.append(kernel_name_);
        CPU_thread_kernel_library_name.append(buf);
        CPU_thread_kernel_library_name.append("-CPU-hand.so");

        std::string command;
        command.append("g++ -O0 -std=c++11 -shared -fpic -o "); // use c++11 for std::atomic
        command.append(CPU_thread_kernel_library_name);
        command.append(" ");
        command.append(CPU_thread_kernel_file_name);
        int ret = system(command.c_str()); // generate dynamic library of CPU kernel thread

        //const char* current_path = std::getenv("PWD");
		char current_path[256];
		getcwd(current_path, 256);
        std::string CPU_thread_kernel_path;
        CPU_thread_kernel_path.append(current_path);
        CPU_thread_kernel_path.append("/");
        CPU_thread_kernel_path.append(CPU_thread_kernel_library_name);
        void* executable = dlopen(CPU_thread_kernel_path.c_str(), RTLD_NOW);
        if (executable == NULL) {
            fprintf(stderr, "executable is NULL\n");
        }

        *(void**)(&onthefly_kernel_cpu_thread_array_[i]) =
            dlsym(executable, "CPU_thread_entry");
        if (onthefly_kernel_cpu_thread_array_ == NULL) {
            fprintf(stderr, "onthefly_kernel_cpu_thread_ is NULL\n");
        }

        onthefly_kernel_cpu_thread_array_params_[i] =
            (unsigned long*)dlsym(executable,
                    "x86_kernel_params");
        if (onthefly_kernel_cpu_thread_array_params_[i]) {
            for (int i = 0; i < armcl_context_->num_total_cpu_cores(); i++) {
                for (int j = 0; j < onthefly_kernel_cpu_thread_num_args_; j++) {
                    fprintf(stdout, "%lu ", onthefly_kernel_cpu_thread_params_[
                            i*onthefly_kernel_cpu_thread_num_args_ + j]);
                }
                fprintf(stdout, "\n");
            }
        }
    }
}

