#!/usr/bin/env python3

from os.path import exists
import os
import sys
import numpy as np
from sklearn import preprocessing
from sklearn import svm
from sklearn.externals import joblib
import re
import subprocess
import copy

number_re = "+([0-9]+\.[01-9]+|[0-9]+)"

def get_values_arr(log, prefix):
    try:
        val = re.findall(prefix+number_re, log)
        return val
    except:
        return "-1.0"

def get_values(log, prefix):
    try:
        val = re.findall(prefix+number_re, log)[0]
        return val
    except:
        return "-1.0"

def get_values_idx(log, prefix, idx):
    try:
        val = re.findall(prefix+number_re, log)[idx]
        return val
    except:
        return "-1.0"


mlpath = "/home/jiyeon/armcl/runtime/ml"
num_total_cores = {
    "skylake":8,
    "kaveri":4
}

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def scanKernel(kernel, workdim, datasize, localsize):
    # workdim, globalsize, localsize
    # read_write_constant, read_write_continuous, read_write_stride, read_write_random, arith_int,arith_float
    kernel_path = os.getcwd() + "/" + kernel + "-kernel.cl"
    compiler_command = mlpath + "/scan " + kernel_path
    eprint(compiler_command)
    output = subprocess.getoutput(compiler_command)

    features = []
    features += [workdim, float(datasize), float(localsize)]
    value = 0
    for opcode in ["read4_constant","write4_constant"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += int(get_values(output, configuration + ": "))
    for opcode in ["read8_constant","write8_constant"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 2*int(get_values(output, configuration + ": "))
    features.append(value)

    value = 0
    for opcode in ["read4_continuous","write4_continuous"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += int(get_values(output, configuration + ": "))
    for opcode in ["read8_continuous","write8_continuous"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 2*int(get_values(output, configuration + ": "))
    features.append(value)

    value = 0
    for opcode in ["read4_stride","write4_stride"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += int(get_values(output, configuration + ": "))
    for opcode in ["read8_stride","write8_stride"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 2*int(get_values(output, configuration + ": "))
    features.append(value)

    value = 0
    for opcode in ["read4_random","write4_random"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += int(get_values(output, configuration + ": "))
    for opcode in ["read8_random","write8_random"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 2*int(get_values(output, configuration + ": "))
    features.append(value)

    value = 0
    for opcode in ["add_int4", "add_int8"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += int(get_values(output, configuration + ": "))
    for opcode in ["mul_int4", "mul_int8"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 2*int(get_values(output, configuration + ": "))
    for opcode in ["div_int4", "div_int8"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 4*int(get_values(output, configuration + ": "))
    features.append(value)

    value = 0
    for opcode in ["add_float4", "add_float8"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += int(get_values(output, configuration + ": "))
    for opcode in ["mul_float4", "mul_float8"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 2*int(get_values(output, configuration + ": "))
    for opcode in ["div_float4", "div_float8"]:
        for loop_level in ["0","1","2"]:
            configuration = kernel + "_loop" + loop_level + "_" + opcode
            value += 4*int(get_values(output, configuration + ": "))
    features.append(value)
    return features

def runModel(features):
    model = joblib.load(mlpath + '/model.pkl')
    X_scaler = joblib.load(mlpath + '/X_scaler.pkl')
    y_scaler = joblib.load(mlpath + '/y_scaler.pkl')
    maxTime = 0

    for cnum in range(1, 45):
        f = copy.deepcopy(features)
        f.append(cnum/9/4.0)
        f.append(cnum%9/8.0)
        print(f)
        f = X_scaler.fit_transform(np.array(f).reshape(1,-1))
        result = model.predict(f)
        if result > maxTime:
            maxTime = result
    return y_scaler.inverse_transform(maxTime)

if __name__ == "__main__":
    kernel = sys.argv[1]    
    workdim = int(sys.argv[2])
    datasize = int(sys.argv[3])
    localsize = int(sys.argv[4])
    arch = sys.argv[5]

    features = scanKernel(kernel, workdim, datasize, localsize)
    for f in features:
        print(int(f), end=" ")
    print("")
    #result = runModel(features)[0]
    #num_cpu = int(round(result)/(num_total_cores[arch]+1))
    #num_gpu = int(round(result)%(num_total_cores[arch]+1))
    #print(str(num_cpu) + " " + str(num_gpu))
