//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLCommandQueue.h"
#include <string.h>
#include <algorithm>
#include <time.h>

int ArmCLCommandQueue::CPUThreading(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "CPUThreading" << std::endl;

    const char* kernel_name = armcl_kernel->kernel_name();
    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();

    struct timespec start, end;
    long long elapsed_time_us;

    std::cout << "total_work_size: " << total_work_size << std::endl;
    std::cout << "local_work_size: " << local_work_size[0] << std::endl;
    std::cout << "num_cpu_compute_units: " << armcl_context_->num_cpu_compute_units() << std::endl;
    //fprintf(stdout, " total number of work groups: %ld\n", total_num_wgs);

    int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

    int (*mbap_cpu_thread_kernel)(int id, void* argp);
    mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();

    int mbap_cpu_thread_kernel_num_args =
        armcl_kernel->mbap_cpu_thread_kernel_num_args();
    unsigned long* mbap_cpu_thread_kernel_params =
        armcl_kernel->mbap_cpu_thread_kernel_params();

    // register reordered map for CPU threads
    for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
        mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
            + mbap_cpu_thread_kernel_num_args - 5] =
            (unsigned long)in_order_map_;
    }

    clock_gettime(CLOCK_REALTIME, &start);

    std::vector<std::thread*> threads;

    size_t alloc_wgs = total_num_wgs / num_cpu_cores;
    size_t alloc_index = 0;

    for (int i = 0; i < num_cpu_cores; i++) {
        //std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
        //    << " alloc_num_wgs: " << alloc_wgs << std::endl;

        if (i == num_cpu_cores-1) {
            // register work item reordered map and
            // work group partitioning information
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)in_order_map_;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 4] =
                (unsigned long)alloc_index;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 3] =
                (unsigned long)local_work_size[0];
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 2] =
                (unsigned long)total_num_wgs - alloc_index;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 1] =
                (unsigned long)(&has_finished[i]);

            std::thread* thread_ = new std::thread(
                    mbap_cpu_thread_kernel,
                    i,
                    (void*)NULL);
            threads.push_back(thread_);
        } else {
            // register work item reordered map and
            // work group partitioning information
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)in_order_map_;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 4] =
                (unsigned long)alloc_index;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 3] =
                (unsigned long)local_work_size[0];
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 2] =
                (unsigned long)alloc_wgs;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 1] =
                (unsigned long)(&has_finished[i]);

            std::thread* thread_ = new std::thread(
                    mbap_cpu_thread_kernel,
                    i,
                    (void*)NULL);
            threads.push_back(thread_);
        }

        alloc_index += alloc_wgs;
    }

    for (int i = 0; i < num_cpu_cores; i++) {
        threads[i]->join();
    }

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;

    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}

int ArmCLCommandQueue::WorkGroupPartitioning_H(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "Work group partitioning (Manual)" << std::endl;
    const char* kernel_name = armcl_kernel->kernel_name();

    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];
    size_t wg_size = local_work_size[0];
    std::cout << "total_work_size: " << total_work_size << std::endl;
    std::cout << "total number of work groups: " << total_num_wgs << std::endl;

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();
    std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
    std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;

    struct timespec start, end;
    long long elapsed_time_us;

    cl_int err;

    int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
    int cpu_ratio = atoi(std::getenv("CPU_RATIO"));
    std::cout << "gpu_ratio: " << gpu_ratio << std::endl;
    std::cout << "cpu_ratio: " << cpu_ratio << std::endl;

    size_t gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
    size_t cpu_wgs = total_num_wgs - gpu_wgs;
    std::cout << "number of work groups for GPU: " << gpu_wgs << std::endl;
    std::cout << "number of work groups for CPU: " << cpu_wgs << std::endl;

    size_t gpu_partitioned_global_work_size[1];
    gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];
    size_t cpu_partitioned_global_work_size[1];
    cpu_partitioned_global_work_size[0] = cpu_wgs * local_work_size[0];

    cl_mem m_in_order_map = clCreateBuffer_p_(context_,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            total_work_size * sizeof(int),
            in_order_map_,
            &err);

    clock_gettime(CLOCK_REALTIME, &start);
    // execution
    if (gpu_wgs) {
        cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
        cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();

        err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1, sizeof(cl_mem), (void*)&m_in_order_map);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: clSetKernelArg for mbap_kernel" << std::endl;
        }

        err = clEnqueueNDRangeKernel_p_(command_queue_,
                mbap_kernel,
                work_dim,
                global_work_offset,
                gpu_partitioned_global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                event);
    }

    if (cpu_wgs) {
        int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

        int (*mbap_cpu_thread_kernel)(int id, void* argp);
        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();

        int mbap_cpu_thread_kernel_num_args =
            armcl_kernel->mbap_cpu_thread_kernel_num_args();
        unsigned long* mbap_cpu_thread_kernel_params =
            armcl_kernel->mbap_cpu_thread_kernel_params();

        // register reordered map for CPU threads
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)in_order_map_;
        }

        std::vector<std::thread*> threads;

        size_t alloc_wgs = cpu_wgs / num_cpu_cores;
        size_t alloc_index = gpu_wgs;

        for (int i = 0; i < num_cpu_cores; i++) {
            std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
                << " alloc_num_wgs: " << alloc_wgs << std::endl;

            if (i == num_cpu_cores-1) {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)in_order_map_;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)total_num_wgs - alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            } else {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)in_order_map_;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)alloc_wgs;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            }

            alloc_index += alloc_wgs;
        }

        for (int i = 0; i < num_cpu_cores; i++) {
            threads[i]->join();
        }
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL MBAP: "
            << "clFinish error"
            << std::endl;
    }

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}

int ArmCLCommandQueue::WorkGroupPartitioning_L(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "Work group partitioning (Manual)" << std::endl;
    const char* kernel_name = armcl_kernel->kernel_name();

    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];
    size_t wg_size = local_work_size[0];
    std::cout << "total_work_size: " << total_work_size << std::endl;
    std::cout << "total number of work groups: " << total_num_wgs << std::endl;

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();
    std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
    std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;

    struct timespec start, end;
    long long elapsed_time_us;

    cl_int err;

    int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
    int cpu_ratio = atoi(std::getenv("CPU_RATIO"));
    std::cout << "gpu_ratio: " << gpu_ratio << std::endl;
    std::cout << "cpu_ratio: " << cpu_ratio << std::endl;

    size_t gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
    size_t cpu_wgs = total_num_wgs - gpu_wgs;
    std::cout << "number of work groups for GPU: " << gpu_wgs << std::endl;
    std::cout << "number of work groups for CPU: " << cpu_wgs << std::endl;

    size_t gpu_partitioned_global_work_size[1];
    //gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];
    gpu_partitioned_global_work_size[0] = global_work_size[0] - (cpu_wgs * local_work_size[0]);
    size_t cpu_partitioned_global_work_size[1];
    cpu_partitioned_global_work_size[0] = cpu_wgs * local_work_size[0];

    size_t gpu_partitioned_index_offset[1];
    gpu_partitioned_index_offset[0] = cpu_wgs * local_work_size[0];

    cl_mem m_in_order_map = clCreateBuffer_p_(context_,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            total_work_size * sizeof(int),
            in_order_map_,
            &err);

    clock_gettime(CLOCK_REALTIME, &start);
    // execution
    if (gpu_wgs) {
        cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
        cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();

        err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1, sizeof(cl_mem), (void*)&m_in_order_map);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: clSetKernelArg for mbap_kernel" << std::endl;
        }

        err = clEnqueueNDRangeKernel_p_(command_queue_,
                mbap_kernel,
                work_dim,
                //global_work_offset,
                gpu_partitioned_index_offset,
                gpu_partitioned_global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                event);
    }

    if (cpu_wgs) {
        int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

        int (*mbap_cpu_thread_kernel)(int id, void* argp);
        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();

        int mbap_cpu_thread_kernel_num_args =
            armcl_kernel->mbap_cpu_thread_kernel_num_args();
        unsigned long* mbap_cpu_thread_kernel_params =
            armcl_kernel->mbap_cpu_thread_kernel_params();

        // register reordered map for CPU threads
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)in_order_map_;
        }

        std::vector<std::thread*> threads;

        size_t alloc_wgs = cpu_wgs / num_cpu_cores;
        size_t alloc_index = 0;

        for (int i = 0; i < num_cpu_cores; i++) {
            std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
                << " alloc_num_wgs: " << alloc_wgs << std::endl;

            if (i == num_cpu_cores-1) {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)in_order_map_;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)cpu_wgs - alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            } else {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)in_order_map_;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)alloc_wgs;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            }

            alloc_index += alloc_wgs;
        }

        for (int i = 0; i < num_cpu_cores; i++) {
            threads[i]->join();
        }
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL MBAP: "
            << "clFinish error"
            << std::endl;
    }

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}

#define SHIFT_H
int ArmCLCommandQueue::FinePar_Manual_Static(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "FinePar method: manual threshold selection; static partitioning" << std::endl;
    const char* kernel_name = armcl_kernel->kernel_name();

    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];
    size_t wg_size = local_work_size[0];
    std::cout << "total_work_size: " << total_work_size << std::endl;
    std::cout << "total number of work groups: " << total_num_wgs << std::endl;

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();
    std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
    std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;

    struct timespec start, end;
    long long elapsed_time_us;

    cl_int err;

    int* finepar_remap = armcl_kernel->finepar_remap_;

    if (!armcl_kernel->is_finepar_partitioned_) {
        int (*mbap_cpu_profiler_kernel)(int id, void* argp);
        mbap_cpu_profiler_kernel = armcl_kernel->mbap_cpu_profiler_kernel();

        int mbap_cpu_profiler_kernel_num_args =
            armcl_kernel->mbap_cpu_profiler_kernel_num_args();
        unsigned long* mbap_cpu_profiler_kernel_params =
            armcl_kernel->mbap_cpu_profiler_kernel_params();

        size_t size_loop_size = total_work_size * sizeof(int);
        int* loop_size = (int*)aligned_alloc(4096, size_loop_size);
        memset(loop_size, 0x00, size_loop_size);

        clock_gettime(CLOCK_REALTIME, &start);
        {
            size_t remaining_wgs = total_num_wgs;
            size_t alloc_wg = remaining_wgs / num_cpu_cores;
            size_t alloc_index = 0;
            int* has_finished_ = (int*)malloc(num_cpu_cores * sizeof(int));
            int suml = 0;

            for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                mbap_cpu_profiler_kernel_params[i*mbap_cpu_profiler_kernel_num_args
                    + mbap_cpu_profiler_kernel_num_args - 5] =
                    (unsigned long)loop_size;
            }

            std::vector<std::thread*> profiler_threads;
            for (int i = 0; i < num_cpu_cores; i++) {
                mbap_cpu_profiler_kernel_params[
                    i * mbap_cpu_profiler_kernel_num_args
                    + mbap_cpu_profiler_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_profiler_kernel_params[
                    i * mbap_cpu_profiler_kernel_num_args
                    + mbap_cpu_profiler_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                if (i == num_cpu_cores - 1) {
                    mbap_cpu_profiler_kernel_params[
                        i * mbap_cpu_profiler_kernel_num_args
                        + mbap_cpu_profiler_kernel_num_args - 2] =
                        (unsigned long)remaining_wgs;
                } else {
                    mbap_cpu_profiler_kernel_params[
                        i * mbap_cpu_profiler_kernel_num_args
                        + mbap_cpu_profiler_kernel_num_args - 2] =
                        (unsigned long)alloc_wg;
                }
                mbap_cpu_profiler_kernel_params[
                    i * mbap_cpu_profiler_kernel_num_args
                    + mbap_cpu_profiler_kernel_num_args - 1]=
                    (unsigned long)(&has_finished_[i]);

                std::thread* thread__ = new std::thread(
                        mbap_cpu_profiler_kernel,
                        i,
                        (void*)NULL);
                profiler_threads.push_back(thread__);
                alloc_index += alloc_wg;
                remaining_wgs -= alloc_wg;
            }

            for (int i = 0; i < num_cpu_cores; i++) {
                profiler_threads[i]->join();
            }

            for (int i = 0; i < num_cpu_cores; i++) {
                suml += has_finished_[i];
            }

            printf("profiling in CPU sum: %d average: %d\n", suml, (int)(suml/total_work_size));
        }
        clock_gettime(CLOCK_REALTIME, &end);

        elapsed_time_us =
            ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
        std::cout << kernel_name << " profiler_run overhead: " << elapsed_time_us << std::endl;

        int* sorted_work_items = (int*)malloc(total_work_size * sizeof(int));
        for (int i = 0; i< total_work_size; i++)
            sorted_work_items[i] = i;

        clock_gettime(CLOCK_REALTIME, &start);
        {
#ifdef SHIFT_H
            // Sorting based work item re-shaping
            std::sort(sorted_work_items, sorted_work_items + total_work_size,
                    [&](int x, int y) {
                    return loop_size[x] < loop_size[y];
                    });
#else
#ifdef SHIFT_L
            // Sorting based work item re-shaping
            std::sort(sorted_work_items, sorted_work_items + total_work_size,
                    [&](int x, int y) {
                    return loop_size[x] > loop_size[y];
                    });
#endif
#endif
        }
        clock_gettime(CLOCK_REALTIME, &end);

        elapsed_time_us =
            ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
        std::cout << kernel_name << " sort-based re-shaping overhead: "
            << elapsed_time_us << std::endl;

        int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
        int cpu_ratio = atoi(std::getenv("CPU_RATIO"));
        std::cout << "gpu_ratio: " << gpu_ratio << std::endl;
        std::cout << "cpu_ratio: " << cpu_ratio << std::endl;

        size_t gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
        size_t cpu_wgs = total_num_wgs - gpu_wgs;
        std::cout << "number of work groups for GPU: " << gpu_wgs << std::endl;
        std::cout << "number of work groups for CPU: " << cpu_wgs << std::endl;

        armcl_kernel->finepar_gpu_wgs_ = gpu_wgs;
        armcl_kernel->finepar_cpu_wgs_ = cpu_wgs;

        int* array_cpu = (int*)calloc(total_work_size, sizeof(int));
        int* array_gpu = (int*)calloc(total_work_size, sizeof(int));
        clock_gettime(CLOCK_REALTIME, &start);
        {
            int cur_array_cpu = 0;
            int cur_array_gpu = 0;
            int index_ = (int)((double)total_work_size * (double)gpu_ratio / (double)100.0);
            int threshold = loop_size[sorted_work_items[index_]];
            std::cout << "index: " << index_ << std::endl;
            std::cout << "threshold: " << threshold << std::endl;

#ifdef SHIFT_H
            for (int i = 0; i < total_work_size; i++) {
                if (loop_size[i] <= threshold) {
                    array_gpu[cur_array_gpu++] = i;
                } else {
                    array_cpu[cur_array_cpu++] = i;
                }
            }
#else
#ifdef SHIFT_L
            for (int i = 0; i < total_work_size; i++) {
                if (loop_size[i] >= threshold) {
                    array_gpu[cur_array_gpu++] = i;
                } else {
                    array_cpu[cur_array_cpu++] = i;
                }
            }
#endif
#endif
            int current = 0;
            for (int i = 0; i < cur_array_gpu; i++)
                finepar_remap[current++] = array_gpu[i];
            for (int i = 0; i < cur_array_cpu; i++)
                finepar_remap[current++] = array_cpu[i];

        }
        clock_gettime(CLOCK_REALTIME, &end);
        elapsed_time_us =
            ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
        std::cout << kernel_name << " threshold-based re-shaping overhead: " << elapsed_time_us << std::endl;

        armcl_kernel->is_finepar_partitioned_ = 1;

        free(array_cpu);
        free(array_gpu);
    }

    //{
    //    static int step = 0;

    //    char file_name1[40] = { 0, };
    //    sprintf(file_name1, "work_items_%d.csv", step);
    //    FILE* out1 = fopen(file_name1, "w");
    //    for (int i = 0; i < wg_size; i++) {
    //        for (int j = i; j < total_work_size; j+=wg_size) {
    //            fprintf(out1, " %d ", loop_size[j]);
    //        }
    //        fprintf(out1, "\n");
    //    }
    //    fclose(out1);

    //    char file_name2[40] = { 0, };
    //    sprintf(file_name2, "sorted_work_items_%d.csv", step);
    //    FILE* out2 = fopen(file_name2, "w");
    //    for (int i = 0; i < wg_size; i++) {
    //        for (int j = i; j < total_work_size; j+=wg_size) {
    //            fprintf(out2, " %d ", loop_size[sorted_work_items[j]]);
    //        }
    //        fprintf(out2, "\n");
    //    }
    //    fclose(out2);

    //    char file_name3[40] = { 0, };

    //    sprintf(file_name3, "shifted_work_items_%d.csv", step++);
    //    FILE* out3 = fopen(file_name3, "w");
    //    for (int i = 0; i < wg_size; i++) {
    //        for (int j = i; j < total_work_size; j+=wg_size) {
    //            fprintf(out3, " %d ", loop_size[remap[j]]);
    //        }
    //        fprintf(out3, "\n");
    //    }
    //    fclose(out3);
    //}

    size_t gpu_partitioned_global_work_size[1];
    gpu_partitioned_global_work_size[0] = armcl_kernel->finepar_gpu_wgs_ * local_work_size[0];
    size_t cpu_partitioned_global_work_size[1];
    cpu_partitioned_global_work_size[0] = armcl_kernel->finepar_cpu_wgs_ * local_work_size[0];

    cl_mem m_remap = clCreateBuffer_p_(context_,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            total_work_size * sizeof(int),
            finepar_remap,
            &err);

    clock_gettime(CLOCK_REALTIME, &start);
    // execution
    if (armcl_kernel->finepar_gpu_wgs_) {
        cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
        cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();

        err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1, sizeof(cl_mem), (void*)&m_remap);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: clSetKernelArg for mbap_kernel" << std::endl;
        }

        err = clEnqueueNDRangeKernel_p_(command_queue_,
                mbap_kernel,
                work_dim,
                global_work_offset,
                gpu_partitioned_global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                event);
    }

    if (armcl_kernel->finepar_cpu_wgs_) {
        int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

        int (*mbap_cpu_thread_kernel)(int id, void* argp);
        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();

        int mbap_cpu_thread_kernel_num_args =
            armcl_kernel->mbap_cpu_thread_kernel_num_args();
        unsigned long* mbap_cpu_thread_kernel_params =
            armcl_kernel->mbap_cpu_thread_kernel_params();

        // register reordered map for CPU threads
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)finepar_remap;
        }

        std::vector<std::thread*> threads;

        size_t alloc_wgs = armcl_kernel->finepar_cpu_wgs_ / num_cpu_cores;
        size_t alloc_index = armcl_kernel->finepar_gpu_wgs_;

        for (int i = 0; i < num_cpu_cores; i++) {
            std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
                << " alloc_num_wgs: " << alloc_wgs << std::endl;

            if (i == num_cpu_cores-1) {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)finepar_remap;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)total_num_wgs - alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            } else {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)finepar_remap;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)alloc_wgs;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            }

            alloc_index += alloc_wgs;
        }

        for (int i = 0; i < num_cpu_cores; i++) {
            threads[i]->join();
        }
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL MBAP: "
            << "clFinish error"
            << std::endl;
    }

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}


////#define KAVERI
//#define SKYLAKE
int ArmCLCommandQueue::OnTheFly(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "OnTheFly" << std::endl;

    const char* kernel_name = armcl_kernel->kernel_name();

    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];
    size_t wg_size = local_work_size[0];

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();

    std::cout << "total_work_size: " << total_work_size << std::endl;
    std::cout << "total number of work groups: " << total_num_wgs << std::endl;
    std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
    std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;

    struct timespec start, end;
    long long elapsed_time_us;

    int mbap_cpu_thread_kernel_num_args =
        armcl_kernel->mbap_cpu_thread_kernel_num_args();

    int (*mbap_cpu_profiler_kernel)(int id, void* argp);
    mbap_cpu_profiler_kernel = armcl_kernel->mbap_cpu_profiler_kernel();

    int mbap_cpu_profiler_kernel_num_args =
        armcl_kernel->mbap_cpu_profiler_kernel_num_args();
    unsigned long* mbap_cpu_profiler_kernel_params =
        armcl_kernel->mbap_cpu_profiler_kernel_params();

    cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
    cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();
    cl_int err;

    cl_mem m_remap = clCreateBuffer_p_(context_,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            total_work_size * sizeof(int),
            remap_,
            &err);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: "
            << "clCreateBuffer (m_remap) error for mbap_kernel"
            << std::endl;
    }

    cl_mem m_in_order_map = clCreateBuffer_p_(context_,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            total_work_size * sizeof(int),
            in_order_map_,
            &err);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: "
            << "clCreateBuffer (m_remap) error for mbap_kernel"
            << std::endl;
    }

    err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1,
            sizeof(cl_mem), (void*)&m_remap);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: "
            << "clSetKernelArg (m_remap) error for mbap_kernel"
            << std::endl;
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: clFinish" << std::endl;
    }

    memset(loop_size_, 0x00, sizeof(int) * total_work_size);
    memset(remap_, 0x00, sizeof(int) * total_work_size);

    int thread_id = 0;
    int num_wgs = (int)(total_num_wgs / 20);
    if (num_wgs > 4096) num_wgs = 4096;
    if (num_wgs < 256) num_wgs = 256;
    int sampling_size = num_wgs * local_work_size[0];

    double adjustment = 1.0;

    /* Profiler run */
    int suml;
    clock_gettime(CLOCK_REALTIME, &start);
    {
        size_t remaining_wgs = num_wgs;
        size_t alloc_wg = remaining_wgs / num_cpu_cores;
        size_t alloc_index = 0;
        int* has_finished_ = (int*)malloc(num_cpu_cores * sizeof(int));

        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            mbap_cpu_profiler_kernel_params[i*mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 5] =
                (unsigned long)loop_size_;
        }

        std::vector<std::thread*> profiler_threads;
        for (int i = 0; i < num_cpu_cores; i++) {
            mbap_cpu_profiler_kernel_params[
                i * mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 4] =
                (unsigned long)alloc_index;
            mbap_cpu_profiler_kernel_params[
                i * mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 3] =
                (unsigned long)local_work_size[0];
            if (i == num_cpu_cores - 1) {
                mbap_cpu_profiler_kernel_params[
                    i * mbap_cpu_profiler_kernel_num_args
                    + mbap_cpu_profiler_kernel_num_args - 2] =
                    (unsigned long)remaining_wgs;
            } else {
                mbap_cpu_profiler_kernel_params[
                    i * mbap_cpu_profiler_kernel_num_args
                    + mbap_cpu_profiler_kernel_num_args - 2] =
                    (unsigned long)alloc_wg;
            }
            mbap_cpu_profiler_kernel_params[
                i * mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 1]=
                (unsigned long)(&has_finished_[i]);

            std::thread* thread__ = new std::thread(
                    mbap_cpu_profiler_kernel,
                    i,
                    (void*)NULL);
            profiler_threads.push_back(thread__);
            alloc_index += alloc_wg;
            remaining_wgs -= alloc_wg;
        }

        for (int i = 0; i < num_cpu_cores; i++) {
            profiler_threads[i]->join();
        }

        for (int i = 0; i < num_cpu_cores; i++)
            suml += has_finished_[i];
    }
    clock_gettime(CLOCK_REALTIME, &end);

    std::cout << "suml: " << suml << std::endl;

    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " profiler_run prologue overhead: " << elapsed_time_us << std::endl;

    /* Reshaping prologue */
    clock_gettime(CLOCK_REALTIME, &start);
    int average = (int)(suml/sampling_size);
    int threshold = (int)((double)average * adjustment);

    int cur_array_cpu = 0;
    int cur_array_gpu = 0;
    for (int i = 0; i < sampling_size; i++) {
        if (loop_size_[i] < threshold) {
            array_gpu_[cur_array_gpu++] = i;
        } else {
            array_cpu_[cur_array_cpu++] = i;
        }
    }

    int gpu_wgs = cur_array_gpu / local_work_size[0];
    int cpu_wgs = num_wgs - gpu_wgs;
    if (cpu_wgs < 16) cpu_wgs = 16;

    int current = 0;
    for (int i = 0; i < cur_array_gpu; i++)
        remap_[current++] = array_gpu_[i];
    for (int i = 0; i < cur_array_cpu; i++)
        remap_[current++] = array_cpu_[i];

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " threshold-based re-shaping prologue overhead: " <<
        elapsed_time_us << std::endl;

    fprintf(stdout, "number of work groups for GPU: %d\n", gpu_wgs);
    fprintf(stdout, "number of work groups for CPU: %d\n", cpu_wgs);

    /* partitioning information */
    size_t gpu_offset[1] = { 0 };
    size_t cpu_offset[1];
    cpu_offset[0] = gpu_wgs * local_work_size[0];
    size_t gpu_partitioned_global_work_size[1];
    gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];
    size_t allocated_wgs = num_wgs;
    size_t allocated_work_items = allocated_wgs * local_work_size[0];
    size_t remaining_wgs = total_num_wgs - allocated_wgs;

    int step = 0;

    int gpu_finished_step = 0;
    int cpu_finished_step = 0;

    int maximum_ = 8192*10;
    int minimum_ = 256;
    int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

    cl_event gpu_kernel_event[100];
    int gpu_run[100] = { 0, };

    /* pipelining */
    clock_gettime(CLOCK_REALTIME, &start);

    std::vector<std::thread*> threads;
    while (true) {
        printf("step: %d\n", step);
        if (gpu_wgs) {
            gpu_run[step] = 1;
            err = clEnqueueNDRangeKernel_p_(command_queue_,
                    mbap_kernel,
                    work_dim,
                    gpu_offset,
                    gpu_partitioned_global_work_size,
                    local_work_size,
                    num_events_in_wait_list,
                    event_wait_list,
                    &gpu_kernel_event[step]);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << "clEnqueueNDRangeKernel error for mbap_kernel"
                    << std::endl;
            }
        }

        int threads_run = 0;
        if (cpu_wgs) {
            size_t alloc_wgs = cpu_wgs / num_cpu_cores;
            size_t alloc_index = cpu_offset[0] / local_work_size[0];
            threads_run = 1;

            for (int i = 0; i < num_cpu_cores; i++) {
                std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
                    << " alloc_num_wgs: " << alloc_wgs << std::endl;

                if (i == num_cpu_cores-1) {
                    int (*mbap_cpu_thread_kernel)(int id, void* argp);
                    mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel_array(step);
                    unsigned long* mbap_cpu_thread_kernel_params =
                        armcl_kernel->mbap_cpu_thread_kernel_array_params(step);

                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 5] =
                        (unsigned long)remap_;
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 4] =
                        (unsigned long)alloc_index;
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 3] =
                        (unsigned long)local_work_size[0];
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 2] =
                        (unsigned long)(allocated_wgs - alloc_index);
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 1] =
                        (unsigned long)(&has_finished[i]);

                    std::thread* thread_ = new std::thread(
                            mbap_cpu_thread_kernel,
                            i,
                            (void*)NULL);

                    threads.push_back(thread_);
                } else {
                    int (*mbap_cpu_thread_kernel)(int id, void* argp);
                    mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel_array(step);
                    unsigned long* mbap_cpu_thread_kernel_params =
                        armcl_kernel->mbap_cpu_thread_kernel_array_params(step);

                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 5] =
                        (unsigned long)remap_;
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 4] =
                        (unsigned long)alloc_index;
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 3] =
                        (unsigned long)local_work_size[0];
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 2] =
                        (unsigned long)alloc_wgs;
                    mbap_cpu_thread_kernel_params[
                        i*mbap_cpu_thread_kernel_num_args
                        + mbap_cpu_thread_kernel_num_args - 1] =
                        (unsigned long)(&has_finished[i]);

                    std::thread* thread_ = new std::thread(
                            mbap_cpu_thread_kernel,
                            i,
                            (void*)NULL);
                    threads.push_back(thread_);
                }

                alloc_index += alloc_wgs;
            }
        }

        /* work item re-shaping pipelining */

        // determine number of work groups to be allocated for next job
        step += 1;
        num_wgs = (int)((double)total_num_wgs / (15.0));
        if (num_wgs > maximum_) num_wgs = maximum_;
        if (num_wgs < minimum_) num_wgs = minimum_;
        if (remaining_wgs < num_wgs * 2)
            num_wgs = remaining_wgs;
        if (num_wgs > 0) {
            sampling_size = num_wgs * local_work_size[0];

            // determine the threshold
            int suml = 0;
            mbap_cpu_profiler_kernel_params[
                0 * mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 4] =
                (unsigned long)(allocated_wgs);
            mbap_cpu_profiler_kernel_params[
                0 * mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 3] =
                (unsigned long)local_work_size[0];
            mbap_cpu_profiler_kernel_params[
                0 * mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 2] =
                (unsigned long)num_wgs;
            mbap_cpu_profiler_kernel_params[
                0 * mbap_cpu_profiler_kernel_num_args
                + mbap_cpu_profiler_kernel_num_args - 1]=
                (unsigned long)(&suml);
            mbap_cpu_profiler_kernel(0, NULL);

            printf("loop sum: %d\n", suml);
            average = suml / sampling_size;
            threshold = (int)((double)average * adjustment);
            //if (threshold < 1) threshold = 1;
            //if (threshold > 50) threshold = 50;

            std::cout << "threshold: " << threshold << std::endl;
            std::cout << "adjustment: " << adjustment << std::endl;

            // threshold-based work-item re-shaping (O(N))
            cur_array_gpu = 0;
            cur_array_cpu = 0;

            for (int i = allocated_work_items;
                    i < allocated_work_items + sampling_size;
                    i++) {
                if (loop_size_[i] < threshold)
                    array_gpu_[cur_array_gpu++] = i;
                else
                    array_cpu_[cur_array_cpu++] = i;
            }

            for (int i = 0; i < cur_array_gpu; i++)
                remap_[current++] = array_gpu_[i];
            for (int i = 0; i < cur_array_cpu; i++)
                remap_[current++] = array_cpu_[i];

            //// compute partitioning information
            gpu_wgs = cur_array_gpu / local_work_size[0];
            cpu_wgs = num_wgs - gpu_wgs;
            if (cpu_wgs < 16) cpu_wgs = 16;

            printf("current: %d\n", current);
            printf("cur_wgs: %d\n", (int)(current/local_work_size[0]));
            printf("num_wgs: %d\n", num_wgs);
            printf("gpu_wgs: %d\n", gpu_wgs);
            printf("cpu_wgs: %d\n", cpu_wgs);

            gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];

            gpu_offset[0] = allocated_wgs * local_work_size[0];
            cpu_offset[0] = gpu_offset[0] + gpu_wgs * local_work_size[0];

            allocated_wgs += num_wgs;
            allocated_work_items = allocated_wgs * local_work_size[0];
            remaining_wgs = total_num_wgs - allocated_wgs;
        } else {
            break;
        }

        {
            cpu_finished_step = has_finished[0]; // currently finished step in the CPU

            if (cpu_finished_step == step && cpu_finished_step - gpu_finished_step > 3) {
                size_t alloc_wgs = (remaining_wgs + num_wgs)/num_cpu_cores;
                size_t alloc_index = gpu_offset[0] / local_work_size[0];

                for (int i = 0; i < num_cpu_cores; i++) {
                    if (i == num_cpu_cores - 1) {
                        int (*mbap_cpu_thread_kernel)(int id, void* argp);
                        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel_array(step);
                        unsigned long* mbap_cpu_thread_kernel_params =
                            armcl_kernel->mbap_cpu_thread_kernel_array_params(step);

                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 5] =
                                (unsigned long)in_order_map_;
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 4] =
                                (unsigned long)alloc_index;
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 3] =
                                (unsigned long)local_work_size[0];
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 2] =
                                (unsigned long)(total_num_wgs - alloc_index);
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 1] =
                                (unsigned long)(&has_finished[i]);

                        std::thread* thread_ = new std::thread(
                                mbap_cpu_thread_kernel,
                                i,
                                (void*)NULL);

                        threads.push_back(thread_);
                    } else {
                        int (*mbap_cpu_thread_kernel)(int id, void* argp);
                        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel_array(step);
                        unsigned long* mbap_cpu_thread_kernel_params =
                            armcl_kernel->mbap_cpu_thread_kernel_array_params(step);

                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 5] =
                                (unsigned long)in_order_map_;
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 4] =
                                (unsigned long)alloc_index;
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 3] =
                                (unsigned long)local_work_size[0];
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 2] =
                                (unsigned long)alloc_wgs;
                        mbap_cpu_thread_kernel_params[
                            i*mbap_cpu_thread_kernel_num_args
                                + mbap_cpu_thread_kernel_num_args - 1] =
                                (unsigned long)(&has_finished[i]);

                        std::thread* thread_ = new std::thread(
                                mbap_cpu_thread_kernel,
                                i,
                                (void*)NULL);
                        threads.push_back(thread_);
                    }
                    alloc_index += alloc_wgs;
                }

                std::cout << "CPU only and finishes" << std::endl;
                break;
            }
        
            if (cpu_finished_step == step) { // cpu too fast
                adjustment *= 0.25;
            }

            if (step > 1 && gpu_finished_step < step - 2 && cpu_finished_step < step - 2) {
                std::cout << "we need to see their speed; wait for a while" << std::endl;

                int gpu_looking = -1;
                for (int i = gpu_finished_step; i < step - 2; i++) {
                    if (gpu_run[i])
                        gpu_looking = i;
                }

                if (gpu_looking == -1) {
                    adjustment *= 1.5;
                } else {
                    clWaitForEvents(1, &gpu_kernel_event[gpu_looking]);
                    gpu_finished_step = step - 2;
                    cpu_finished_step = has_finished[0];

                    if (cpu_finished_step == step) {
                        adjustment *= 0.5;
                    } else if (cpu_finished_step > gpu_finished_step) {
                        adjustment *= 0.8;
                    } else if (cpu_finished_step == gpu_finished_step) {
                        adjustment *= 1.0;
                    } else if (cpu_finished_step < gpu_finished_step) {
                        adjustment *= 1.5;
                    }
                }
            }

            std::cout << "next_step: " << step << std::endl;
            std::cout << "gpu_finished_step: " << gpu_finished_step << std::endl;
            std::cout << "cpu_finished_step: " << cpu_finished_step << std::endl;
        }

        //if (adjustment < 0.0125) adjustment = 0.0125;

//        // check both CPU/GPU engines already have finished.
//        {
//            cl_int gpu_kernel_execution_status;
//            if (has_finished[0] == step) {
//                err = clGetEventInfo_p_(gpu_kernel_event,
//                        CL_EVENT_COMMAND_EXECUTION_STATUS,
//                        sizeof(cl_int),
//                        &gpu_kernel_execution_status,
//                        NULL);
//                if (err != CL_SUCCESS) {
//                    std::cout << "ArmCL OnTheFly: clGetEventInfo error" << std::endl;
//                }
//
//                if (gpu_kernel_execution_status == CL_COMPLETE) {
//                    next_job += 1;
//                }
//            } else if (step < 5) {
//                while (true) {
//                    if (has_finished[0] == step) {
//                        // CPU has finished eariler
//                        adjustment *= 0.5;
//                        break;
//                    } else {
//                        err = clGetEventInfo_p_(gpu_kernel_event,
//                                CL_EVENT_COMMAND_EXECUTION_STATUS,
//                                sizeof(cl_int),
//                                &gpu_kernel_execution_status,
//                                NULL);
//                        if (err != CL_SUCCESS) {
//                            std::cout << "ArmCL OnTheFly: clGetEventInfo error" << std::endl;
//                        }
//
//                        if (gpu_kernel_execution_status == CL_COMPLETE) {
//                            std::cout << "GPU has finished eariler" << std::endl;
//                            adjustment *= 1.5;
//                            next_job += 1;
//                            break;
//                        }
//                    }
//                }
//            }
//        }
    }

    for (int i = 0; i < threads.size(); i++) {
        threads[i]->join();
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL MBAP: "
            << "clFinish error"
            << std::endl;
    }

    free(has_finished);

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;

    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}

////#define SORT_H
////#define SORT_L
//
//int ArmCLCommandQueue::Optimal(ArmCLKernel* armcl_kernel,
//        cl_uint work_dim,
//        const size_t* global_work_offset,
//        const size_t* global_work_size,
//        const size_t* local_work_size,
//        cl_uint num_events_in_wait_list,
//        const cl_event* event_wait_list,
//        cl_event* event)
//{
//    std::cout << "Optimal (Manual)" << std::endl;
//    const char* kernel_name = armcl_kernel->kernel_name();
//
//    size_t total_work_size = global_work_size[0];
//    size_t total_num_wgs = total_work_size / local_work_size[0];
//    size_t wg_size = local_work_size[0];
//    std::cout << "total_work_size: " << total_work_size << std::endl;
//    std::cout << "total number of work groups: " << total_num_wgs << std::endl;
//
//    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
//    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();
//    std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
//    std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;
//
//    struct timespec start, end;
//    long long elapsed_time_us;
//
//    cl_int err;
//
//    int (*mbap_cpu_profiler_kernel)(int id, void* argp);
//    mbap_cpu_profiler_kernel = armcl_kernel->mbap_cpu_profiler_kernel();
//
//    int mbap_cpu_profiler_kernel_num_args =
//        armcl_kernel->mbap_cpu_profiler_kernel_num_args();
//    unsigned long* mbap_cpu_profiler_kernel_params =
//        armcl_kernel->mbap_cpu_profiler_kernel_params();
//
//    size_t size_loop_size = total_work_size * sizeof(int);
//    int* loop_size = (int*)aligned_alloc(4096, size_loop_size);
//    memset(loop_size, 0x00, size_loop_size);
//
//    err = clFinish(command_queue_);
//    if (err != CL_SUCCESS) {
//        std::cout  << "ArmCL: clFinish" << std::endl;
//    }
//
//    clock_gettime(CLOCK_REALTIME, &start);
//    {
//        size_t remaining_wgs = total_num_wgs;
//        size_t alloc_wg = remaining_wgs / num_cpu_cores;
//        size_t alloc_index = 0;
//        int* has_finished_ = (int*)malloc(num_cpu_cores * sizeof(int));
//        //int sum = 0;
//
//        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
//            mbap_cpu_profiler_kernel_params[i*mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 5] =
//                (unsigned long)loop_size;
//        }
//
//        std::vector<std::thread*> profiler_threads;
//        for (int i = 0; i < num_cpu_cores; i++) {
//            mbap_cpu_profiler_kernel_params[
//                i * mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 4] =
//                (unsigned long)alloc_index;
//            mbap_cpu_profiler_kernel_params[
//                i * mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 3] =
//                (unsigned long)local_work_size[0];
//            if (i == num_cpu_cores - 1) {
//                mbap_cpu_profiler_kernel_params[
//                    i * mbap_cpu_profiler_kernel_num_args
//                    + mbap_cpu_profiler_kernel_num_args - 2] =
//                    (unsigned long)remaining_wgs;
//            } else {
//                mbap_cpu_profiler_kernel_params[
//                    i * mbap_cpu_profiler_kernel_num_args
//                    + mbap_cpu_profiler_kernel_num_args - 2] =
//                    (unsigned long)alloc_wg;
//            }
//            mbap_cpu_profiler_kernel_params[
//                i * mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 1]=
//                (unsigned long)(&has_finished_[i]);
//
//            std::thread* thread__ = new std::thread(
//                    mbap_cpu_profiler_kernel,
//                    i,
//                    (void*)NULL);
//            profiler_threads.push_back(thread__);
//            alloc_index += alloc_wg;
//            remaining_wgs -= alloc_wg;
//        }
//
//        for (int i = 0; i < num_cpu_cores; i++) {
//            profiler_threads[i]->join();
//        }
//
//        //printf("profiling in CPU sum: %d average: %d\n", sum, (int)(sum/total_work_size));
//    }
//    clock_gettime(CLOCK_REALTIME, &end);
//
//    elapsed_time_us =
//        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
//    std::cout << kernel_name << " profiler_run overhead: " << elapsed_time_us << std::endl;
//
//    int* sorted_work_items = (int*)malloc(total_work_size * sizeof(int));
//    for (int i = 0; i< total_work_size; i++)
//        sorted_work_items[i] = i;
//
//    clock_gettime(CLOCK_REALTIME, &start);
//    {
//#ifdef SORT_H
//        // Sorting based work item re-shaping
//        std::sort(sorted_work_items, sorted_work_items + total_work_size,
//                [&](int x, int y) {
//                return loop_size[x] < loop_size[y];
//                });
//#else
//#ifdef SORT_L
//        // Sorting based work item re-shaping
//        std::sort(sorted_work_items, sorted_work_items + total_work_size,
//                [&](int x, int y) {
//                return loop_size[x] > loop_size[y];
//                });
//#endif
//#endif
//    }
//    clock_gettime(CLOCK_REALTIME, &end);
//
//    elapsed_time_us =
//        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
//    std::cout << kernel_name << " sort-based re-shaping overhead: "
//        << elapsed_time_us << std::endl;
//
//    int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
//    int cpu_ratio = atoi(std::getenv("CPU_RATIO"));
//    std::cout << "gpu_ratio: " << gpu_ratio << std::endl;
//    std::cout << "cpu_ratio: " << cpu_ratio << std::endl;
//
//    size_t gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
//    size_t cpu_wgs = total_num_wgs - gpu_wgs;
//    std::cout << "number of work groups for GPU: " << gpu_wgs << std::endl;
//    std::cout << "number of work groups for CPU: " << cpu_wgs << std::endl;
//
//    size_t gpu_partitioned_global_work_size[1];
//    gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];
//    size_t cpu_partitioned_global_work_size[1];
//    cpu_partitioned_global_work_size[0] = cpu_wgs * local_work_size[0];
//
//    int* remap = (int*)aligned_alloc(4096, total_work_size * sizeof(int));
//    for (int i = 0; i < total_work_size; i++)
//        remap[i] = sorted_work_items[i];
//
//    {
//        static int step = 0;
//
//        char file_name1[40] = { 0, };
//        sprintf(file_name1, "work_items_%d.csv", step);
//        FILE* out1 = fopen(file_name1, "w");
//        for (int i = 0; i < wg_size; i++) {
//            for (int j = i; j < total_work_size; j+=wg_size) {
//                fprintf(out1, " %d ", loop_size[j]);
//            }
//            fprintf(out1, "\n");
//        }
//        fclose(out1);
//
//        char file_name2[40] = { 0, };
//        sprintf(file_name2, "sorted_work_items_%d.csv", step);
//        FILE* out2 = fopen(file_name2, "w");
//        for (int i = 0; i < wg_size; i++) {
//            for (int j = i; j < total_work_size; j+=wg_size) {
//                fprintf(out2, " %d ", loop_size[sorted_work_items[j]]);
//            }
//            fprintf(out2, "\n");
//        }
//        fclose(out2);
//
//        char file_name3[40] = { 0, };
//
//        sprintf(file_name3, "shifted_work_items_%d.csv", step++);
//        FILE* out3 = fopen(file_name3, "w");
//        for (int i = 0; i < wg_size; i++) {
//            for (int j = i; j < total_work_size; j+=wg_size) {
//                fprintf(out3, " %d ", loop_size[remap[j]]);
//            }
//            fprintf(out3, "\n");
//        }
//        fclose(out3);
//    }
//
//    cl_mem m_remap = clCreateBuffer_p_(context_,
//            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
//            total_work_size * sizeof(int),
//            remap,
//            &err);
//
//    clock_gettime(CLOCK_REALTIME, &start);
//    // execution
//    if (gpu_wgs) {
//        cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
//        cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();
//
//        err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1, sizeof(cl_mem), (void*)&m_remap);
//        if (err != CL_SUCCESS) {
//            std::cout << "ArmCL: clSetKernelArg for mbap_kernel" << std::endl;
//        }
//
//        err = clEnqueueNDRangeKernel_p_(command_queue_,
//                mbap_kernel,
//                work_dim,
//                global_work_offset,
//                gpu_partitioned_global_work_size,
//                local_work_size,
//                num_events_in_wait_list,
//                event_wait_list,
//                event);
//    }
//
//    if (cpu_wgs) {
//        int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));
//
//        int (*mbap_cpu_thread_kernel)(int id, void* argp);
//        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();
//
//        int mbap_cpu_thread_kernel_num_args =
//            armcl_kernel->mbap_cpu_thread_kernel_num_args();
//        unsigned long* mbap_cpu_thread_kernel_params =
//            armcl_kernel->mbap_cpu_thread_kernel_params();
//
//        // register reordered map for CPU threads
//        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
//            mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
//                + mbap_cpu_thread_kernel_num_args - 5] =
//                (unsigned long)remap;
//        }
//
//        std::vector<std::thread*> threads;
//
//        size_t alloc_wgs = cpu_wgs / num_cpu_cores;
//        size_t alloc_index = gpu_wgs;
//
//        for (int i = 0; i < num_cpu_cores; i++) {
//            std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
//                << " alloc_num_wgs: " << alloc_wgs << std::endl;
//
//            if (i == num_cpu_cores-1) {
//                // register work item reordered map and
//                // work group partitioning information
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 5] =
//                    (unsigned long)remap;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 4] =
//                    (unsigned long)alloc_index;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 3] =
//                    (unsigned long)local_work_size[0];
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 2] =
//                    (unsigned long)total_num_wgs - alloc_index;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 1] =
//                    (unsigned long)(&has_finished[i]);
//
//                std::thread* thread_ = new std::thread(
//                        mbap_cpu_thread_kernel,
//                        i,
//                        (void*)NULL);
//                threads.push_back(thread_);
//            } else {
//                // register work item reordered map and
//                // work group partitioning information
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 5] =
//                    (unsigned long)remap;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 4] =
//                    (unsigned long)alloc_index;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 3] =
//                    (unsigned long)local_work_size[0];
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 2] =
//                    (unsigned long)alloc_wgs;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 1] =
//                    (unsigned long)(&has_finished[i]);
//
//                std::thread* thread_ = new std::thread(
//                        mbap_cpu_thread_kernel,
//                        i,
//                        (void*)NULL);
//                threads.push_back(thread_);
//            }
//
//            alloc_index += alloc_wgs;
//        }
//
//        for (int i = 0; i < num_cpu_cores; i++) {
//            threads[i]->join();
//        }
//    }
//
//    err = clFinish(command_queue_);
//    if (err != CL_SUCCESS) {
//        std::cout << "ArmCL MBAP: "
//            << "clFinish error"
//            << std::endl;
//    }
//
//    clock_gettime(CLOCK_REALTIME, &end);
//    elapsed_time_us =
//        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
//    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;
//
//    return CL_SUCCESS;
//}

#define SHIFT_H
int ArmCLCommandQueue::Optimal(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "Optimal (Manual)" << std::endl;
    const char* kernel_name = armcl_kernel->kernel_name();

    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];
    size_t wg_size = local_work_size[0];
    std::cout << "total_work_size: " << total_work_size << std::endl;
    std::cout << "total number of work groups: " << total_num_wgs << std::endl;

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();
    std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
    std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;

    struct timespec start, end;
    long long elapsed_time_us;

    cl_int err;

    cl_kernel mbap_profiler_kernel = armcl_kernel->mbap_profiler_kernel();
    cl_uint mbap_profiler_kernel_num_args =
        armcl_kernel->mbap_profiler_kernel_num_args();

    int (*mbap_cpu_profiler_kernel)(int id, void* argp);
    mbap_cpu_profiler_kernel = armcl_kernel->mbap_cpu_profiler_kernel();

    int mbap_cpu_profiler_kernel_num_args =
        armcl_kernel->mbap_cpu_profiler_kernel_num_args();
    unsigned long* mbap_cpu_profiler_kernel_params =
        armcl_kernel->mbap_cpu_profiler_kernel_params();

    size_t size_loop_size = total_work_size * sizeof(int);
    int* loop_size = (int*)aligned_alloc(4096, size_loop_size);
    memset(loop_size, 0x00, size_loop_size);

    cl_mem m_loop_size = clCreateBuffer_p_(context_,
            CL_MEM_WRITE_ONLY|CL_MEM_USE_HOST_PTR,
            size_loop_size,
            loop_size,
            &err);
    if (err != CL_SUCCESS) {
        std::cout << "clCreateBuffer failed for loop profile data"
            << std::endl;
        return -1;
    }

    std::cout << "mbap_profiler_kernel_num_args: " << mbap_profiler_kernel_num_args << std::endl;

    err = clSetKernelArg_p_(mbap_profiler_kernel,
            mbap_profiler_kernel_num_args - 1,
            sizeof(cl_mem),
            (void*)&m_loop_size);
    if (err != CL_SUCCESS) {
        std::cout << "clSetKernelArg failed for mbap profile data"
            << std::endl;
        return -1;
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout  << "ArmCL: clFinish" << std::endl;
    }

    clock_gettime(CLOCK_REALTIME, &start);
    {
        cl_event initial_profiler_kernel_event;
        err = clEnqueueNDRangeKernel_p_(command_queue_,
                mbap_profiler_kernel,
                work_dim,
                global_work_offset,
                global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                event);
        if (err != CL_SUCCESS) {
            std::cout << "clEnqueueNDRangeKernel failed for loop profile data: "
                << " errcode: " << err
                << std::endl;
            return -1;
        }

        //err = clWaitForEvents(1, &initial_profiler_kernel_event);
        err = clFinish(command_queue_);

        //size_t remaining_wgs = total_num_wgs;
        //size_t alloc_wg = remaining_wgs / num_cpu_cores;
        //size_t alloc_index = 0;
        //int* has_finished_ = (int*)malloc(num_cpu_cores * sizeof(int));
        ////int sum = 0;

        //for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
        //    mbap_cpu_profiler_kernel_params[i*mbap_cpu_profiler_kernel_num_args
        //        + mbap_cpu_profiler_kernel_num_args - 5] =
        //        (unsigned long)loop_size;
        //}

        //std::vector<std::thread*> profiler_threads;
        //for (int i = 0; i < num_cpu_cores; i++) {
        //    mbap_cpu_profiler_kernel_params[
        //        i * mbap_cpu_profiler_kernel_num_args
        //        + mbap_cpu_profiler_kernel_num_args - 4] =
        //        (unsigned long)alloc_index;
        //    mbap_cpu_profiler_kernel_params[
        //        i * mbap_cpu_profiler_kernel_num_args
        //        + mbap_cpu_profiler_kernel_num_args - 3] =
        //        (unsigned long)local_work_size[0];
        //    if (i == num_cpu_cores - 1) {
        //        mbap_cpu_profiler_kernel_params[
        //            i * mbap_cpu_profiler_kernel_num_args
        //            + mbap_cpu_profiler_kernel_num_args - 2] =
        //            (unsigned long)remaining_wgs;
        //    } else {
        //        mbap_cpu_profiler_kernel_params[
        //            i * mbap_cpu_profiler_kernel_num_args
        //            + mbap_cpu_profiler_kernel_num_args - 2] =
        //            (unsigned long)alloc_wg;
        //    }
        //    mbap_cpu_profiler_kernel_params[
        //        i * mbap_cpu_profiler_kernel_num_args
        //        + mbap_cpu_profiler_kernel_num_args - 1]=
        //        (unsigned long)(&has_finished_[i]);

        //    std::thread* thread__ = new std::thread(
        //            mbap_cpu_profiler_kernel,
        //            i,
        //            (void*)NULL);
        //    profiler_threads.push_back(thread__);
        //    alloc_index += alloc_wg;
        //    remaining_wgs -= alloc_wg;
        //}

        //for (int i = 0; i < num_cpu_cores; i++) {
        //    profiler_threads[i]->join();
        //}

        //printf("profiling in CPU sum: %d average: %d\n", sum, (int)(sum/total_work_size));
    }
    clock_gettime(CLOCK_REALTIME, &end);

    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " profiler_run overhead: " << elapsed_time_us << std::endl;

    int* sorted_work_items = (int*)malloc(total_work_size * sizeof(int));
    for (int i = 0; i< total_work_size; i++)
        sorted_work_items[i] = i;

    clock_gettime(CLOCK_REALTIME, &start);
    {
#ifdef SHIFT_H
        // Sorting based work item re-shaping
        std::sort(sorted_work_items, sorted_work_items + total_work_size,
                [&](int x, int y) {
                return loop_size[x] < loop_size[y];
                });
#else
#ifdef SHIFT_L
        // Sorting based work item re-shaping
        std::sort(sorted_work_items, sorted_work_items + total_work_size,
                [&](int x, int y) {
                return loop_size[x] > loop_size[y];
                });
#endif
#endif
    }
    clock_gettime(CLOCK_REALTIME, &end);

    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " sort-based re-shaping overhead: "
        << elapsed_time_us << std::endl;

    int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
    int cpu_ratio = atoi(std::getenv("CPU_RATIO"));
    std::cout << "gpu_ratio: " << gpu_ratio << std::endl;
    std::cout << "cpu_ratio: " << cpu_ratio << std::endl;

    size_t gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
    size_t cpu_wgs = total_num_wgs - gpu_wgs;
    std::cout << "number of work groups for GPU: " << gpu_wgs << std::endl;
    std::cout << "number of work groups for CPU: " << cpu_wgs << std::endl;

    size_t gpu_partitioned_global_work_size[1];
    gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];
    size_t cpu_partitioned_global_work_size[1];
    cpu_partitioned_global_work_size[0] = cpu_wgs * local_work_size[0];

    int* remap = (int*)malloc(total_work_size * sizeof(int));
    int* array_cpu = (int*)calloc(total_work_size, sizeof(int));
    int* array_gpu = (int*)calloc(total_work_size, sizeof(int));
    clock_gettime(CLOCK_REALTIME, &start);
    {
        int cur_array_cpu = 0;
        int cur_array_gpu = 0;
        int index_ = (int)((double)total_work_size * (double)gpu_ratio / (double)100.0);
        int threshold = loop_size[sorted_work_items[index_]];
        std::cout << "index: " << index_ << std::endl;
        std::cout << "threshold: " << threshold << std::endl;

#ifdef SHIFT_H
        for (int i = 0; i < total_work_size; i++) {
            if (loop_size[i] <= threshold) {
                array_gpu[cur_array_gpu++] = i;
            } else {
                array_cpu[cur_array_cpu++] = i;
            }
        }
#else
#ifdef SHIFT_L
        for (int i = 0; i < total_work_size; i++) {
            if (loop_size[i] >= threshold) {
                array_gpu[cur_array_gpu++] = i;
            } else {
                array_cpu[cur_array_cpu++] = i;
            }
        }
#endif
#endif
        int current = 0;
        for (int i = 0; i < cur_array_gpu; i++)
            remap[current++] = array_gpu[i];
        for (int i = 0; i < cur_array_cpu; i++)
            remap[current++] = array_cpu[i];

    }
    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " threshold-based re-shaping overhead: " << elapsed_time_us << std::endl;

    free(array_gpu);
    free(array_cpu);

    {
        static int step = 0;

        char file_name1[40] = { 0, };
        sprintf(file_name1, "work_items_%d.csv", step);
        FILE* out1 = fopen(file_name1, "w");
        for (int i = 0; i < wg_size; i++) {
            for (int j = i; j < total_work_size; j+=wg_size) {
                fprintf(out1, " %d ", loop_size[j]);
            }
            fprintf(out1, "\n");
        }
        fclose(out1);

        char file_name2[40] = { 0, };
        sprintf(file_name2, "sorted_work_items_%d.csv", step);
        FILE* out2 = fopen(file_name2, "w");
        for (int i = 0; i < wg_size; i++) {
            for (int j = i; j < total_work_size; j+=wg_size) {
                fprintf(out2, " %d ", loop_size[sorted_work_items[j]]);
            }
            fprintf(out2, "\n");
        }
        fclose(out2);

        char file_name3[40] = { 0, };

        sprintf(file_name3, "shifted_work_items_%d.csv", step++);
        FILE* out3 = fopen(file_name3, "w");
        for (int i = 0; i < wg_size; i++) {
            for (int j = i; j < total_work_size; j+=wg_size) {
                fprintf(out3, " %d ", loop_size[remap[j]]);
            }
            fprintf(out3, "\n");
        }
        fclose(out3);
    }

    cl_mem m_remap = clCreateBuffer_p_(context_,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            total_work_size * sizeof(int),
            remap,
            &err);

    clock_gettime(CLOCK_REALTIME, &start);
    // execution
    if (gpu_wgs) {
        cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
        cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();

        err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1, sizeof(cl_mem), (void*)&m_remap);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: clSetKernelArg for mbap_kernel" << std::endl;
        }

        err = clEnqueueNDRangeKernel_p_(command_queue_,
                mbap_kernel,
                work_dim,
                global_work_offset,
                gpu_partitioned_global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                event);
    }

    if (cpu_wgs) {
        int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

        int (*mbap_cpu_thread_kernel)(int id, void* argp);
        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();

        int mbap_cpu_thread_kernel_num_args =
            armcl_kernel->mbap_cpu_thread_kernel_num_args();
        unsigned long* mbap_cpu_thread_kernel_params =
            armcl_kernel->mbap_cpu_thread_kernel_params();

        // register reordered map for CPU threads
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)remap;
        }

        std::vector<std::thread*> threads;

        size_t alloc_wgs = cpu_wgs / num_cpu_cores;
        size_t alloc_index = gpu_wgs;

        for (int i = 0; i < num_cpu_cores; i++) {
            std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
                << " alloc_num_wgs: " << alloc_wgs << std::endl;

            if (i == num_cpu_cores-1) {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)remap;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)total_num_wgs - alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            } else {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)remap;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)alloc_wgs;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            }

            alloc_index += alloc_wgs;
        }

        for (int i = 0; i < num_cpu_cores; i++) {
            threads[i]->join();
        }
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL MBAP: "
            << "clFinish error"
            << std::endl;
    }

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}

//#define SHIFT_H
////#define SHIFT_L
//
//int ArmCLCommandQueue::Optimal(ArmCLKernel* armcl_kernel,
//        cl_uint work_dim,
//        const size_t* global_work_offset,
//        const size_t* global_work_size,
//        const size_t* local_work_size,
//        cl_uint num_events_in_wait_list,
//        const cl_event* event_wait_list,
//        cl_event* event)
//{
//    std::cout << "Optimal (Manual)" << std::endl;
//    const char* kernel_name = armcl_kernel->kernel_name();
//
//    size_t total_work_size = global_work_size[0];
//    size_t total_num_wgs = total_work_size / local_work_size[0];
//    size_t wg_size = local_work_size[0];
//    std::cout << "total_work_size: " << total_work_size << std::endl;
//    std::cout << "total number of work groups: " << total_num_wgs << std::endl;
//
//    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
//    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();
//    std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
//    std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;
//
//    struct timespec start, end;
//    long long elapsed_time_us;
//
//    cl_int err;
//
//    int (*mbap_cpu_profiler_kernel)(int id, void* argp);
//    mbap_cpu_profiler_kernel = armcl_kernel->mbap_cpu_profiler_kernel();
//
//    int mbap_cpu_profiler_kernel_num_args =
//        armcl_kernel->mbap_cpu_profiler_kernel_num_args();
//    unsigned long* mbap_cpu_profiler_kernel_params =
//        armcl_kernel->mbap_cpu_profiler_kernel_params();
//
//    size_t size_loop_size = total_work_size * sizeof(int);
//    int* loop_size = (int*)aligned_alloc(4096, size_loop_size);
//    memset(loop_size, 0x00, size_loop_size);
//
//    err = clFinish(command_queue_);
//    if (err != CL_SUCCESS) {
//        std::cout  << "ArmCL: clFinish" << std::endl;
//    }
//
//    clock_gettime(CLOCK_REALTIME, &start);
//    {
//        size_t remaining_wgs = total_num_wgs;
//        size_t alloc_wg = remaining_wgs / num_cpu_cores;
//        size_t alloc_index = 0;
//        int* has_finished_ = (int*)malloc(num_cpu_cores * sizeof(int));
//        //int sum = 0;
//
//        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
//            mbap_cpu_profiler_kernel_params[i*mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 5] =
//                (unsigned long)loop_size;
//        }
//
//        std::vector<std::thread*> profiler_threads;
//        for (int i = 0; i < num_cpu_cores; i++) {
//            mbap_cpu_profiler_kernel_params[
//                i * mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 4] =
//                (unsigned long)alloc_index;
//            mbap_cpu_profiler_kernel_params[
//                i * mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 3] =
//                (unsigned long)local_work_size[0];
//            if (i == num_cpu_cores - 1) {
//                mbap_cpu_profiler_kernel_params[
//                    i * mbap_cpu_profiler_kernel_num_args
//                    + mbap_cpu_profiler_kernel_num_args - 2] =
//                    (unsigned long)remaining_wgs;
//            } else {
//                mbap_cpu_profiler_kernel_params[
//                    i * mbap_cpu_profiler_kernel_num_args
//                    + mbap_cpu_profiler_kernel_num_args - 2] =
//                    (unsigned long)alloc_wg;
//            }
//            mbap_cpu_profiler_kernel_params[
//                i * mbap_cpu_profiler_kernel_num_args
//                + mbap_cpu_profiler_kernel_num_args - 1]=
//                (unsigned long)(&has_finished_[i]);
//
//            std::thread* thread__ = new std::thread(
//                    mbap_cpu_profiler_kernel,
//                    i,
//                    (void*)NULL);
//            profiler_threads.push_back(thread__);
//            alloc_index += alloc_wg;
//            remaining_wgs -= alloc_wg;
//        }
//
//        for (int i = 0; i < num_cpu_cores; i++) {
//            profiler_threads[i]->join();
//        }
//
//        //printf("profiling in CPU sum: %d average: %d\n", sum, (int)(sum/total_work_size));
//    }
//    clock_gettime(CLOCK_REALTIME, &end);
//
//    elapsed_time_us =
//        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
//    std::cout << kernel_name << " profiler_run overhead: " << elapsed_time_us << std::endl;
//
//    int* sorted_work_items = (int*)malloc(total_work_size * sizeof(int));
//    for (int i = 0; i< total_work_size; i++)
//        sorted_work_items[i] = i;
//
//    clock_gettime(CLOCK_REALTIME, &start);
//    {
//#ifdef SHIFT_H
//        // Sorting based work item re-shaping
//        std::sort(sorted_work_items, sorted_work_items + total_work_size,
//                [&](int x, int y) {
//                return loop_size[x] < loop_size[y];
//                });
//#else
//#ifdef SHIFT_L
//        // Sorting based work item re-shaping
//        std::sort(sorted_work_items, sorted_work_items + total_work_size,
//                [&](int x, int y) {
//                return loop_size[x] > loop_size[y];
//                });
//#endif
//#endif
//    }
//    clock_gettime(CLOCK_REALTIME, &end);
//
//    elapsed_time_us =
//        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
//    std::cout << kernel_name << " sort-based re-shaping overhead: "
//        << elapsed_time_us << std::endl;
//
//    int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
//    int cpu_ratio = atoi(std::getenv("CPU_RATIO"));
//    std::cout << "gpu_ratio: " << gpu_ratio << std::endl;
//    std::cout << "cpu_ratio: " << cpu_ratio << std::endl;
//
//    size_t gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
//    size_t cpu_wgs = total_num_wgs - gpu_wgs;
//    std::cout << "number of work groups for GPU: " << gpu_wgs << std::endl;
//    std::cout << "number of work groups for CPU: " << cpu_wgs << std::endl;
//
//    size_t gpu_partitioned_global_work_size[1];
//    gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];
//    size_t cpu_partitioned_global_work_size[1];
//    cpu_partitioned_global_work_size[0] = cpu_wgs * local_work_size[0];
//
//    int* remap = (int*)malloc(total_work_size * sizeof(int));
//    int* array_cpu = (int*)calloc(total_work_size, sizeof(int));
//    int* array_gpu = (int*)calloc(total_work_size, sizeof(int));
//    clock_gettime(CLOCK_REALTIME, &start);
//    {
//        int cur_array_cpu = 0;
//        int cur_array_gpu = 0;
//        int index_ = (int)((double)total_work_size * (double)gpu_ratio / (double)100.0);
//        int threshold = loop_size[sorted_work_items[index_]];
//        std::cout << "index: " << index_ << std::endl;
//        std::cout << "threshold: " << threshold << std::endl;
//
//#ifdef SHIFT_H
//        for (int i = 0; i < total_work_size; i++) {
//            if (loop_size[i] <= threshold) {
//                array_gpu[cur_array_gpu++] = i;
//            } else {
//                array_cpu[cur_array_cpu++] = i;
//            }
//        }
//#else
//#ifdef SHIFT_L
//        for (int i = 0; i < total_work_size; i++) {
//            if (loop_size[i] >= threshold) {
//                array_gpu[cur_array_gpu++] = i;
//            } else {
//                array_cpu[cur_array_cpu++] = i;
//            }
//        }
//#endif
//#endif
//        int current = 0;
//        for (int i = 0; i < cur_array_gpu; i++)
//            remap[current++] = array_gpu[i];
//        for (int i = 0; i < cur_array_cpu; i++)
//            remap[current++] = array_cpu[i];
//
//    }
//    clock_gettime(CLOCK_REALTIME, &end);
//    elapsed_time_us =
//        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
//    std::cout << kernel_name << " threshold-based re-shaping overhead: " << elapsed_time_us << std::endl;
//
//    free(array_gpu);
//    free(array_cpu);
//
//    {
//        static int step = 0;
//
//        char file_name1[40] = { 0, };
//        sprintf(file_name1, "work_items_%d.csv", step);
//        FILE* out1 = fopen(file_name1, "w");
//        for (int i = 0; i < wg_size; i++) {
//            for (int j = i; j < total_work_size; j+=wg_size) {
//                fprintf(out1, " %d ", loop_size[j]);
//            }
//            fprintf(out1, "\n");
//        }
//        fclose(out1);
//
//        char file_name2[40] = { 0, };
//        sprintf(file_name2, "sorted_work_items_%d.csv", step);
//        FILE* out2 = fopen(file_name2, "w");
//        for (int i = 0; i < wg_size; i++) {
//            for (int j = i; j < total_work_size; j+=wg_size) {
//                fprintf(out2, " %d ", loop_size[sorted_work_items[j]]);
//            }
//            fprintf(out2, "\n");
//        }
//        fclose(out2);
//
//        char file_name3[40] = { 0, };
//
//        sprintf(file_name3, "shifted_work_items_%d.csv", step++);
//        FILE* out3 = fopen(file_name3, "w");
//        for (int i = 0; i < wg_size; i++) {
//            for (int j = i; j < total_work_size; j+=wg_size) {
//                fprintf(out3, " %d ", loop_size[remap[j]]);
//            }
//            fprintf(out3, "\n");
//        }
//        fclose(out3);
//    }
//
//    cl_mem m_remap = clCreateBuffer_p_(context_,
//            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
//            total_work_size * sizeof(int),
//            remap,
//            &err);
//
//    clock_gettime(CLOCK_REALTIME, &start);
//    // execution
//    if (gpu_wgs) {
//        cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
//        cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();
//
//        err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1, sizeof(cl_mem), (void*)&m_remap);
//        if (err != CL_SUCCESS) {
//            std::cout << "ArmCL: clSetKernelArg for mbap_kernel" << std::endl;
//        }
//
//        err = clEnqueueNDRangeKernel_p_(command_queue_,
//                mbap_kernel,
//                work_dim,
//                global_work_offset,
//                gpu_partitioned_global_work_size,
//                local_work_size,
//                num_events_in_wait_list,
//                event_wait_list,
//                event);
//    }
//
//    if (cpu_wgs) {
//        int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));
//
//        int (*mbap_cpu_thread_kernel)(int id, void* argp);
//        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();
//
//        int mbap_cpu_thread_kernel_num_args =
//            armcl_kernel->mbap_cpu_thread_kernel_num_args();
//        unsigned long* mbap_cpu_thread_kernel_params =
//            armcl_kernel->mbap_cpu_thread_kernel_params();
//
//        // register reordered map for CPU threads
//        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
//            mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
//                + mbap_cpu_thread_kernel_num_args - 5] =
//                (unsigned long)remap;
//        }
//
//        std::vector<std::thread*> threads;
//
//        size_t alloc_wgs = cpu_wgs / num_cpu_cores;
//        size_t alloc_index = gpu_wgs;
//
//        for (int i = 0; i < num_cpu_cores; i++) {
//            std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
//                << " alloc_num_wgs: " << alloc_wgs << std::endl;
//
//            if (i == num_cpu_cores-1) {
//                // register work item reordered map and
//                // work group partitioning information
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 5] =
//                    (unsigned long)remap;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 4] =
//                    (unsigned long)alloc_index;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 3] =
//                    (unsigned long)local_work_size[0];
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 2] =
//                    (unsigned long)total_num_wgs - alloc_index;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 1] =
//                    (unsigned long)(&has_finished[i]);
//
//                std::thread* thread_ = new std::thread(
//                        mbap_cpu_thread_kernel,
//                        i,
//                        (void*)NULL);
//                threads.push_back(thread_);
//            } else {
//                // register work item reordered map and
//                // work group partitioning information
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 5] =
//                    (unsigned long)remap;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 4] =
//                    (unsigned long)alloc_index;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 3] =
//                    (unsigned long)local_work_size[0];
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 2] =
//                    (unsigned long)alloc_wgs;
//                mbap_cpu_thread_kernel_params[
//                    i*mbap_cpu_thread_kernel_num_args
//                    + mbap_cpu_thread_kernel_num_args - 1] =
//                    (unsigned long)(&has_finished[i]);
//
//                std::thread* thread_ = new std::thread(
//                        mbap_cpu_thread_kernel,
//                        i,
//                        (void*)NULL);
//                threads.push_back(thread_);
//            }
//
//            alloc_index += alloc_wgs;
//        }
//
//        for (int i = 0; i < num_cpu_cores; i++) {
//            threads[i]->join();
//        }
//    }
//
//    err = clFinish(command_queue_);
//    if (err != CL_SUCCESS) {
//        std::cout << "ArmCL MBAP: "
//            << "clFinish error"
//            << std::endl;
//    }
//
//    clock_gettime(CLOCK_REALTIME, &end);
//    elapsed_time_us =
//        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
//    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;
//
//    return CL_SUCCESS;
//}
//
