/// Profiler
/// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#include "profiler.h"
#include <sys/time.h>

static long long getTimestamp(void)
{
    static __thread struct timeval tv;
    gettimeofday(&tv, NULL);
    return 1000000LL * tv.tv_sec + tv.tv_usec;
}

using namespace std;

//------------------------------------------------------------------------------
// MCTProfiler
MCTProfiler::MCTProfiler(int num_memories)
{
    _pmu_manager = new PMUReqManager();

    int num_events = 5;
    _attr_mapping = (std::pair<__u32, __u64>*)
        malloc(num_events * sizeof(std::pair<__u32, __u64>));

    _attr_mapping[0] = make_pair<__u32, __u64>
        (TOTAL_CYCLES_TYPE, TOTAL_CYCLES_CONFIG);
    _attr_mapping[1] = make_pair<__u32, __u64>
        (LLC_MISSES_TYPE, LLC_MISSES_CONFIG);
    _attr_mapping[2] = make_pair<__u32, __u64>
        (MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_TYPE,
         MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_CONFIG);
    _attr_mapping[3] = make_pair<__u32, __u64>
        (MEMORY_CONTROLLER_REQUESTS_PREFETCH_REQUESTS_TO_DCT_TYPE,
         MEMORY_CONTROLLER_REQUESTS_PREFETCH_REQUESTS_TO_DCT_CONFIG);
    _attr_mapping[4] = make_pair<__u32, __u64>
        (MEMORY_CONTROLLER_REQUESTS_WRITE_REQUESTS_TO_DCT_TYPE,
         MEMORY_CONTROLLER_REQUESTS_WRITE_REQUESTS_TO_DCT_CONFIG);

    _attr_type = (event_type*)malloc(num_events * sizeof(event_type));
    _attr_type[0] = PER_CORE;
    _attr_type[1] = PER_CLUSTER;
    _attr_type[2] = PER_CLUSTER;
    _attr_type[3] = PER_CLUSTER;
    _attr_type[4] = PER_CLUSTER;

    _pmu_manager->InitPerfEventAttr(_attr_mapping, _attr_type, num_events);
    _pmu_manager->OpenEventFDs();

    num_memories_ = num_memories;

    cout << "Hello! I'm a profiler" << endl;
}

MCTProfiler::~MCTProfiler()
{
    cout << "Bye~ from the profiler" << endl;

    free(_attr_mapping);
    free(_attr_type);
    _pmu_manager->CloseEventFDs();
}

void MCTProfiler::ProfilerInfo(string path)
{
    return;
}

void MCTProfiler::StartProfiling()
{
    _pmu_manager->GetCurrentCount(_start);

    clock_gettime(CLOCK_REALTIME, &start);
}

void MCTProfiler::StopProfiling(string prefix, string path)
{
    struct timespec stop;
    clock_gettime(CLOCK_REALTIME, &stop);

    const int TOTAL_CYCLES = 0;
    const int LLC_MISSES = 1;
    const int MCT_READ_REQUESTS = 2;
    const int MCT_PREFETCH_REQUESTS = 3;
    const int MCT_WRITE_REQUESTS = 4;
    sc_event_values _end;

    unsigned long* per_thread_llc_misses;
    _pmu_manager->RecordAllCounters();
    _pmu_manager->GetCurrentCount(_end);

    /* don't save profiling info. if PMU reading request is within 10ms */
    if (_end.values[0][0] == _start.values[0][0]) return;

    unsigned long sum_llc_misses = 0;
    unsigned long sum_mct_read_requests = 0;
    unsigned long sum_mct_prefetch_requests = 0;
    unsigned long sum_mct_write_requests = 0;

    for (int i = 0; i < NUM_CORES; i+=NUM_CORES_IN_NODE) {
        unsigned long total_cycles = _end.values[TOTAL_CYCLES][i]
            - _start.values[TOTAL_CYCLES][i];
        unsigned long llc_misses = _end.values[LLC_MISSES][i]
            - _start.values[LLC_MISSES][i];
        unsigned long mct_read_requests = _end.values[MCT_READ_REQUESTS][i]
            - _start.values[MCT_READ_REQUESTS][i];
        unsigned long mct_prefetch_requests = _end.values[MCT_PREFETCH_REQUESTS][i]
            - _start.values[MCT_PREFETCH_REQUESTS][i];
        unsigned long mct_write_requests = _end.values[MCT_WRITE_REQUESTS][i]
            - _start.values[MCT_WRITE_REQUESTS][i];

        mct_read_requests -= mct_prefetch_requests;

        sum_llc_misses += llc_misses;

        //std::cout << "CPU node: " << i/NUM_CORES_IN_NODE
        //    << " total_cycles: " << total_cycles
        //    << " llc_misses: " << llc_misses
        //    << " mct_read_requests: " << mct_read_requests
        //    << " mct_prefetch_requests: " << mct_prefetch_requests
        //    << " mct_write_requests: " << mct_write_requests << std::endl;

        int node_num = i/NUM_CORES_IN_NODE;
        if (node_num < num_memories_) {
            //ofstream out(path.c_str());
            //out.open(path.c_str(), ios_base::out | ios_base::app);
            ofstream out;
            out.open(path.c_str(), ios_base::out | ios_base::app);

            //out << prefix << "_total_cycles: " << total_cycles << std::endl;
            out << prefix << "_node" << node_num << "_llc_misses: " << llc_misses << std::endl;
            out << prefix << "_node" << node_num << "_mct_read_requests: " << mct_read_requests << std::endl;
            out << prefix << "_node" << node_num << "_mct_prefetch_requests: " << mct_prefetch_requests << std::endl;
            out << prefix << "_node" << node_num << "_mct_write_requests: " << mct_write_requests << std::endl;

            sum_mct_read_requests += mct_read_requests;
            sum_mct_prefetch_requests += mct_prefetch_requests;
            sum_mct_write_requests +=mct_write_requests;

            out.close();
        }
    }

    {
        //ofstream out(path.c_str());
        ofstream out;
        out.open(path.c_str(), ios_base::out | ios_base::app);

        double elapsed_time = (stop.tv_sec - start.tv_sec)*1000000000 + (stop.tv_nsec - start.tv_nsec);
        elapsed_time /= 1000;

        out << prefix << "_elapsed_time: " << (unsigned long)elapsed_time << std::endl;
        out << prefix << "_llc_misses: " << sum_llc_misses << std::endl;

        out << prefix << "_sum_mct_read_requests: " << sum_mct_read_requests << std::endl;
        out << prefix << "_sum_mct_prefetch_requests: " << sum_mct_prefetch_requests << std::endl;
        out << prefix << "_sum_mct_write_requests: " << sum_mct_write_requests << std::endl;

        out.close();
    }

    //std::cout << " sum_llc_misses: " << sum_llc_misses
    //    << " sum_mct_read_requests: " << sum_mct_read_requests
    //    << " sum_mct_prefetch_requests: " << sum_mct_prefetch_requests
    //    << " sum_mct_write_requests: " << sum_mct_write_requests << std::endl;
}

//------------------------------------------------------------------------------
// DCTProfiler
DCTProfiler::DCTProfiler(int dct_num, int num_memories)
    : dct_num_(dct_num)
{
    _pmu_manager = new PMUReqManager();

    int num_events = 5;
    _attr_mapping = (std::pair<__u32, __u64>*)
        malloc(num_events * sizeof(std::pair<__u32, __u64>));

    _attr_mapping[0] = make_pair<__u32, __u64>
        (TOTAL_CYCLES_TYPE, TOTAL_CYCLES_CONFIG);
    _attr_mapping[1] = make_pair<__u32, __u64>
        (LLC_MISSES_TYPE, LLC_MISSES_CONFIG);

    if (dct_num_ == 0) {
        _attr_mapping[2] = make_pair<__u32, __u64>
            (DRAM_ACCESSES_DCT0_PAGE_HIT_TYPE,
             DRAM_ACCESSES_DCT0_PAGE_HIT_CONFIG);
        _attr_mapping[3] = make_pair<__u32, __u64>
            (DRAM_ACCESSES_DCT0_PAGE_MISS_TYPE,
             DRAM_ACCESSES_DCT0_PAGE_MISS_CONFIG);
        _attr_mapping[4] = make_pair<__u32, __u64>
            (DRAM_ACCESSES_DCT0_PAGE_CONFLICT_TYPE,
             DRAM_ACCESSES_DCT0_PAGE_CONFLICT_CONFIG);
    } else if (dct_num_ == 1) {
        _attr_mapping[2] = make_pair<__u32, __u64>
            (DRAM_ACCESSES_DCT1_PAGE_HIT_TYPE,
             DRAM_ACCESSES_DCT1_PAGE_HIT_CONFIG);
        _attr_mapping[3] = make_pair<__u32, __u64>
            (DRAM_ACCESSES_DCT1_PAGE_MISS_TYPE,
             DRAM_ACCESSES_DCT1_PAGE_MISS_CONFIG);
        _attr_mapping[4] = make_pair<__u32, __u64>
            (DRAM_ACCESSES_DCT1_PAGE_CONFLICT_TYPE,
             DRAM_ACCESSES_DCT1_PAGE_CONFLICT_CONFIG);
    }

    _attr_type = (event_type*)malloc(num_events * sizeof(event_type));
    _attr_type[0] = PER_CORE;
    _attr_type[1] = PER_CLUSTER;
    _attr_type[2] = PER_CLUSTER;
    _attr_type[3] = PER_CLUSTER;
    _attr_type[4] = PER_CLUSTER;

    _pmu_manager->InitPerfEventAttr(_attr_mapping, _attr_type, num_events);
    _pmu_manager->OpenEventFDs();

    num_memories_ = num_memories;

    cout << "Hello! I'm a profiler" << endl;
}

DCTProfiler::~DCTProfiler()
{
    cout << "Bye~ from the profiler" << endl;

    free(_attr_mapping);
    free(_attr_type);
    _pmu_manager->CloseEventFDs();
}

void DCTProfiler::ProfilerInfo(string path)
{
    return;
}

void DCTProfiler::StartProfiling()
{
    _pmu_manager->GetCurrentCount(_start);
}

void DCTProfiler::StopProfiling(string prefix, string path)
{
    const int TOTAL_CYCLES = 0;
    const int LLC_MISSES = 1;
    const int DCT_PAGE_HITS = 2;
    const int DCT_PAGE_MISSES = 3;
    const int DCT_PAGE_CONFLICTS = 4;
    sc_event_values _end;

    unsigned long* per_thread_llc_misses;
    _pmu_manager->RecordAllCounters();
    _pmu_manager->GetCurrentCount(_end);

    /* don't save profiling info. if PMU reading request is within 10ms */
    if (_end.values[0][0] == _start.values[0][0]) return;

    unsigned long sum_llc_misses = 0;
    unsigned long sum_dct_page_hits = 0;
    unsigned long sum_dct_page_misses = 0;
    unsigned long sum_dct_page_conflicts = 0;

    for (int i = 0; i < NUM_CORES; i+=NUM_CORES_IN_NODE) {
        unsigned long total_cycles = _end.values[TOTAL_CYCLES][i]
            - _start.values[TOTAL_CYCLES][i];
        unsigned long llc_misses = _end.values[LLC_MISSES][i]
            - _start.values[LLC_MISSES][i];
        unsigned long dct_page_hits = _end.values[DCT_PAGE_HITS][i]
            - _start.values[DCT_PAGE_HITS][i];
        unsigned long dct_page_misses = _end.values[DCT_PAGE_MISSES][i]
            - _start.values[DCT_PAGE_MISSES][i];
        unsigned long dct_page_conflicts = _end.values[DCT_PAGE_CONFLICTS][i]
            - _start.values[DCT_PAGE_CONFLICTS][i];

        sum_llc_misses += llc_misses;

        std::cout << "CPU node: " << i/NUM_CORES_IN_NODE
            << " total_cycles: " << total_cycles
            << " llc_misses: " << llc_misses
            << " dct" << dct_num_ << "_page_hits: " << dct_page_hits
            << " dct" << dct_num_ << "_page_misses: " << dct_page_misses
            << " dct" << dct_num_ << "_page_conflicts: " << dct_page_conflicts
            << std::endl;

        int node_num = i/NUM_CORES_IN_NODE;
        if (node_num < num_memories_) {
            ofstream out;
            out.open(path.c_str(), ios_base::out | ios_base::app);

            out << prefix << "_node" << node_num << "_dct" << dct_num_ << "_page_hits: " << dct_page_hits << std::endl;
            out << prefix << "_node" << node_num << "_dct" << dct_num_ << "_page_misses: " << dct_page_misses << std::endl;
            out << prefix << "_node" << node_num << "_dct" << dct_num_ << "_page_conflicts: " << dct_page_conflicts << std::endl;

            sum_dct_page_hits += dct_page_hits;
            sum_dct_page_misses += dct_page_misses;
            sum_dct_page_conflicts += dct_page_conflicts;

            out.close();
        }
    }

    {
        //ofstream out(path.c_str());
        ofstream out;
        out.open(path.c_str(), ios_base::out | ios_base::app);

        out << prefix << "_sum_dct" << dct_num_ << "_page_hits: " << sum_dct_page_hits << std::endl;
        out << prefix << "_sum_dct" << dct_num_ << "_page_misses: " << sum_dct_page_misses << std::endl;
        out << prefix << "_sum_dct" << dct_num_ << "_page_conflicts: " << sum_dct_page_conflicts << std::endl;

        out.close();
    }

    std::cout << " sum_llc_misses: " << sum_llc_misses
        << " sum_dct" << dct_num_ << "_page_hits: " << sum_dct_page_hits
        << " sum_dct" << dct_num_ << "_page_misses: " << sum_dct_page_misses
        << " sum_dct" << dct_num_ << "_page_conflicts: " << sum_dct_page_conflicts
        << std::endl;
}

