// MatAdd: Synthetic workloads (Matrix Addition) based on PolyBench/GPU 1.0 suite.
//
// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#include <cstdlib>
#include <iostream>

#include <CL/cl.h>

#include <thread>
#include <vector>
#include <atomic>
#include <malloc.h>
#include <math.h>

#include "profiler_interface.h"
#include <omp.h>

#define ALPHA 3
#define BETA 2

#define MAX_SOURCE_SIZE (0x100000)

//define the error threshold for the results "not matching"
#define PERCENT_DIFF_ERROR_THRESHOLD 0.05

/* Can switch DATA_TYPE between float and double */
#ifdef FLOAT
typedef float DATA_TYPE;
#else
#ifdef DOUBLE
typedef double DATA_TYPE;
#else
#ifdef INT
typedef int DATA_TYPE;
#endif
#endif
#endif

/* Workload description */
// C[i] += A[i][j] + B[j][i]

/* Workload partitioning */
int num_cpu_cores = 1;
int num_gpu_compute_units = 8;
int throttling_mod = 1;
int throttling_alloc = 1;
int gpu_percent;

/* Problem size */
int NI = 0;
int NJ = 0;
int NK = 0;

/* Amount of computational operations */
int computation = 1;

/* Thread block dimensions */
int DIM_LOCAL_WORK_GROUP_X = 16;
int DIM_LOCAL_WORK_GROUP_Y = 4;
int DIM_LOCAL_WORK_GROUP_Z = 2;
int work_dim = 2;

/* Data buffers */
DATA_TYPE* A;
DATA_TYPE* B;
DATA_TYPE* C;

cl_platform_id platform_id;
cl_device_id device_id;   
cl_uint num_devices;
cl_uint num_platforms;
cl_int errcode;
cl_context clGPUContext;
cl_kernel clKernel;
cl_command_queue clCommandQue;
cl_program clProgram;
cl_mem a_mem_obj;
cl_mem b_mem_obj;
cl_mem c_mem_obj;
FILE *fp;
char *source_str;
size_t source_size;

inline int cpu_first(std::atomic_int* worklist)
{
    int current = worklist->fetch_add(1);
    return current;
}

inline int cpu_next(std::atomic_int* worklist)
{
    int current = worklist->fetch_add(1);
    return current;
}

inline bool cpu_more(int current, int n_tasks)
{
    return current < n_tasks;
}

void Mat2DAdd_CPU_Threads(int id,
                          DATA_TYPE* A,
                          DATA_TYPE* B,
                          DATA_TYPE* C,
                          DATA_TYPE alpha,
                          DATA_TYPE beta,
                          int NI,
                          int NJ,
                          int work_dim,
                          size_t* global_size,
                          size_t* local_size,
                          std::atomic_int* worklist,
                          int num_tasks)
{
    int processed_wgs = 0;

#if WORK_DIM == 1
    for (int wg_id = cpu_first(worklist);
             cpu_more(wg_id, num_tasks);
             wg_id = cpu_next(worklist)) {
        processed_wgs++;

        for (int global_id = wg_id * local_size[0];
                 global_id < wg_id * local_size[0] + local_size[0];
                 global_id++) {
            int i = global_id;

#if COMPUTATION == 0
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
											C[i * (NJ) + j] = A[i * (NJ) + j] +
																		 B[i * (NJ) + j];
                }
            }
#elif COMPUTATION == 1
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
											C[i * (NJ) + j] = alpha * A[i * (NJ) + j] +
																		 beta * B[i * (NJ) + j];
                }
            }
#elif COMPUTATION == 2
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
											C[i * (NJ) + j] = alpha * alpha * A[i * (NJ) + j] +
																		 beta * beta * B[i * (NJ) + j];
                }
            }
#elif COMPUTATION == 4
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
											C[i * (NJ) + j] = alpha * alpha * alpha * alpha * A[i * (NJ) + j] +
																		 beta * beta * beta * beta * B[i * (NJ) + j];
                }
            }
#elif COMPUTATION == 8
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
											C[i * (NJ) + j] = alpha * alpha * alpha * alpha *
																		 alpha * alpha * alpha * alpha * A[i * (NJ) + j] +
																		 beta * beta * beta * beta *
																		 beta * beta * beta * beta * B[i * (NJ) + j];
                }
            }
#elif COMPUTATION == 16
            if (i < NI) {
                for (int j = 0; j < NJ; j++) {
											C[i * (NJ) + j] = alpha * alpha * alpha * alpha *
																		 alpha * alpha * alpha * alpha *
																		 alpha * alpha * alpha * alpha *
																		 alpha * alpha * alpha * alpha * A[i * (NJ) + j] +
																		 beta * beta * beta * beta *
																		 beta * beta * beta * beta *
																		 beta * beta * beta * beta *
																		 beta * beta * beta * beta * B[i * (NJ) + j];
                }
            }
#endif
        }
    }
#endif

    printf("thread%d: %d\n", id, processed_wgs);
}

#define SMALL_FLOAT_VAL 0.00000001f

float absVal(float a)
{
	if (a < 0) {
		return (a * -1);
	} else { 
		return a;
	}
}

float percentDiff(double val1, double val2)
{
	if ((absVal(val1) < 0.01) && (absVal(val2) < 0.01)) {
		return 0.0f;
    } else {
        return 100.0f * (absVal(absVal(val1 - val2) / absVal(val1 + SMALL_FLOAT_VAL)));
	}
} 

void compareResults(DATA_TYPE* C, DATA_TYPE* C_validation)
{
    int i, j, fail;
    fail = 0;

    // Compare C1 and C2
    for (i=0; i < NI; i++) {
	for (j=0; j < NJ; j++) {
	    if (percentDiff(C[i * (NJ) + j],C_validation[i * (NJ) + j])
		    > PERCENT_DIFF_ERROR_THRESHOLD) {
		fail++;
	    }
	}
    }

	// Print results
	printf("Non-Matching CPU-GPU Outputs Beyond Error Threshold of %4.2f Percent: %d\n", PERCENT_DIFF_ERROR_THRESHOLD, fail);
}

void read_cl_file()
{
	// Load the kernel source code into the array source_str
	fp = fopen("kernel-hand.cl", "r");
	if (!fp) {
		fprintf(stderr, "Failed to load kernel.\n");
		exit(1);
	}
	source_str = (char*)malloc(MAX_SOURCE_SIZE);
	source_size = fread( source_str, 1, MAX_SOURCE_SIZE, fp);
	fclose( fp );
}

void init(DATA_TYPE *A, DATA_TYPE *B, DATA_TYPE *C)
{
    #pragma omp parallel for
    for (int i = 0; i < NI; i++) {
        for (int j = 0; j < NJ; j++) {
							A[i * (NJ) + j] = (i * (NJ) + j)%10;
        }
    }

    #pragma omp parallel for
    for (int i = 0; i < NI; i++) {
        for (int j = 0; j < NJ; j++) {
							B[i * (NJ) + j] = (i * (NJ) + j)%9;
        }
    }

    #pragma omp parallel for
    for (int i = 0; i < NI; i++) {
        for (int j = 0; j < NJ; j++) {
							C[i * (NJ) + j] = 0;
        }
    }
}

void cl_initialization()
{	
    char str_temp[1024];

    // Get platform and device information
    errcode = clGetPlatformIDs(1, &platform_id, &num_platforms);
    if(errcode == CL_SUCCESS) printf("number of platforms is %d\n",num_platforms);
    else printf("Error getting platform IDs\n");

    errcode = clGetPlatformInfo(platform_id,CL_PLATFORM_NAME, sizeof(str_temp), str_temp,NULL);
    if(errcode == CL_SUCCESS) printf("platform name is %s\n",str_temp);
    else printf("Error getting platform name\n");

    errcode = clGetPlatformInfo(platform_id, CL_PLATFORM_VERSION, sizeof(str_temp), str_temp,NULL);
    if(errcode == CL_SUCCESS) printf("platform version is %s\n",str_temp);
    else printf("Error getting platform version\n");

    errcode = clGetDeviceIDs( platform_id, CL_DEVICE_TYPE_GPU, 1, &device_id, &num_devices);
    if(errcode == CL_SUCCESS) printf("number of devices is %d\n", num_devices);
    else printf("Error getting device IDs\n");

    errcode = clGetDeviceInfo(device_id,CL_DEVICE_NAME, sizeof(str_temp), str_temp,NULL);
    if(errcode == CL_SUCCESS) printf("device name is %s\n",str_temp);
    else printf("Error getting device name\n");

    // Create an OpenCL context
    clGPUContext = clCreateContext( NULL, 1, &device_id, NULL, NULL, &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating context\n");

    //Create a command-queue
    cl_queue_properties prop[] = { 0 };
    clCommandQue = clCreateCommandQueueWithProperties(clGPUContext, device_id, prop, &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating command queue\n");
}

void cl_mem_init(DATA_TYPE* A, DATA_TYPE* B, DATA_TYPE* C)
{
    a_mem_obj = clCreateBuffer(clGPUContext,
                               CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
                               sizeof(DATA_TYPE) * NI * NJ,
                               A,
                               &errcode);
    b_mem_obj = clCreateBuffer(clGPUContext,
                               CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
                               sizeof(DATA_TYPE) * NI * NJ,
                               B,
                               &errcode);
    c_mem_obj = clCreateBuffer(clGPUContext,
                               CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR,
                               sizeof(DATA_TYPE) * NI * NJ,
                               C,
                               &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating buffers\n");
}

void cl_load_prog()
{
    // Create a program from the kernel source
    clProgram = clCreateProgramWithSource(clGPUContext, 1, (const char **)&source_str, (const size_t *)&source_size, &errcode);

    if(errcode != CL_SUCCESS) printf("Error in creating program\n");

    // Build the program
    char build_option[100];
#if defined(FLOAT)
    sprintf(build_option, "-I. -cl-std=CL2.0 -DFLOAT -DWORK_DIM=%d -DCOMPUTATION=%d", work_dim, computation);
#elif defined(DOUBLE)
    sprintf(build_option, "-I. -cl-std=CL2.0 -DDOUBLE -DWORK_DIM=%d -DCOMPUTATION=%d", work_dim, computation);
#elif defined(INT)
    sprintf(build_option, "-I. -cl-std=CL2.0 -DINT -DWORK_DIM=%d -DCOMPUTATION=%d", work_dim, computation);
#endif

    errcode = clBuildProgram(clProgram, 1, &device_id, build_option, NULL, NULL);
    if(errcode != CL_SUCCESS) printf("Error in building program\n");

    // Create the OpenCL kernel
    clKernel = clCreateKernel(clProgram, "mat2dadd_dynamic_alloc_atomic", &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating kernel\n");
    clFinish(clCommandQue);
}

void cl_launch_kernel()
{
    int ni=NI;
    int nj=NJ;

    DATA_TYPE alpha = ALPHA;
    DATA_TYPE beta = BETA;

#if WORK_DIM == 1
    size_t localWorkSize[1], globalWorkSize[1];
    localWorkSize[0] = DIM_LOCAL_WORK_GROUP_X;
    globalWorkSize[0] = (size_t)ceil(((float)NI) / ((float)DIM_LOCAL_WORK_GROUP_X)) * DIM_LOCAL_WORK_GROUP_X;
#endif

    // Set the arguments of the kernel
    errcode =  clSetKernelArg(clKernel, 0, sizeof(cl_mem), (void*)&a_mem_obj);
    errcode |= clSetKernelArg(clKernel, 1, sizeof(cl_mem), (void*)&b_mem_obj);
    errcode |= clSetKernelArg(clKernel, 2, sizeof(cl_mem), (void*)&c_mem_obj);
    errcode |= clSetKernelArg(clKernel, 3, sizeof(DATA_TYPE), (void*)&alpha);
    errcode |= clSetKernelArg(clKernel, 4, sizeof(DATA_TYPE), (void*)&beta);
    errcode |= clSetKernelArg(clKernel, 5, sizeof(int), (void*)&ni);
    errcode |= clSetKernelArg(clKernel, 6, sizeof(int), (void*)&nj);
    errcode |= clSetKernelArg(clKernel, 7, sizeof(int), (void*)&throttling_mod);
    errcode |= clSetKernelArg(clKernel, 8, sizeof(int), (void*)&throttling_alloc);
    if(errcode != CL_SUCCESS) printf("Error in setting arguments\n");

#if WORK_DIM == 1
    size_t total_num_wgs = globalWorkSize[0] / localWorkSize[0];

    size_t gpu_partitioned_global_work_size[1];
	int num_gpu_wgs = total_num_wgs*gpu_percent/100;
    gpu_partitioned_global_work_size[0] = num_gpu_wgs * localWorkSize[0];

    std::atomic_int* worklist = (std::atomic_int*)malloc(sizeof(std::atomic_int));
    worklist[0].store(num_gpu_wgs);
    int n_tasks = total_num_wgs;

    StopProfiling("prelude", "profiling.txt");
    StartProfiling();

	if(gpu_percent != 0) {
		errcode = clEnqueueNDRangeKernel(clCommandQue, clKernel, work_dim, NULL,
										 gpu_partitioned_global_work_size, localWorkSize,
										 0, NULL, NULL);
		if (errcode != CL_SUCCESS) printf("Error in launching kernel\n");

		clFlush(clCommandQue);
	}

	if(gpu_percent != 100) {
		std::vector<std::thread*> threads;

		for (int i = 0; i < num_cpu_cores; i++) {
			std::thread* thread_ = new std::thread(Mat2DAdd_CPU_Threads, i,
					A, B, C, ALPHA, BETA, NI, NJ,
					work_dim, globalWorkSize, localWorkSize,
					worklist, n_tasks);
			threads.push_back(thread_);
		}

		for (int i = 0; i < threads.size(); i++) {
			threads[i]->join();
		}
	}

    clFinish(clCommandQue);

    StopProfiling("mat2dadd", "profiling.txt");

#endif
}

void cl_clean_up()
{
    // Clean up
    errcode = clFlush(clCommandQue);
    errcode = clFinish(clCommandQue);
    errcode = clReleaseKernel(clKernel);
    errcode = clReleaseProgram(clProgram);
    errcode = clReleaseMemObject(a_mem_obj);
    errcode = clReleaseMemObject(b_mem_obj);
    errcode = clReleaseMemObject(c_mem_obj);
    errcode = clReleaseCommandQueue(clCommandQue);
    errcode = clReleaseContext(clGPUContext);
    if(errcode != CL_SUCCESS) printf("Error in cleanup\n");
}

void mat2dadd(DATA_TYPE *A, DATA_TYPE *B, DATA_TYPE *C)
{
#if COMPUTATION == 0
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
                C[i * (NJ) + j] = A[i * (NJ) + j] +
                               B[i * (NJ) + j];
        }
    }
#elif COMPUTATION == 1
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
					C[i * (NJ) + j] = ALPHA * A[i * (NJ) + j] +
						BETA * B[i * (NJ) + j];
        }
    }
#elif COMPUTATION == 2
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
					C[i * (NJ) + j] = ALPHA * ALPHA * A[i * (NJ) + j] +
						BETA * BETA * B[i * (NJ) + j];
        }
    }
#elif COMPUTATION == 4
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
					C[i * (NJ) + j] = ALPHA * ALPHA * ALPHA * ALPHA * A[i * (NJ) + j] +
						BETA * BETA * BETA * BETA * B[i * (NJ) + j];
        }
    }
#elif COMPUTATION == 8
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
					C[i * (NJ) + j] = ALPHA * ALPHA * ALPHA * ALPHA *
						ALPHA * ALPHA * ALPHA * ALPHA * A[i * (NJ) + j] +
						BETA * BETA * BETA * BETA *
						BETA * BETA * BETA * BETA * B[i * (NJ) + j];
        }
    }
#elif COMPUTATION == 16
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
					C[i * (NJ) + j] = ALPHA * ALPHA * ALPHA * ALPHA *
						ALPHA * ALPHA * ALPHA * ALPHA *
						ALPHA * ALPHA * ALPHA * ALPHA *
						ALPHA * ALPHA * ALPHA * ALPHA * A[i * (NJ) + j] +
						BETA * BETA * BETA * BETA *
						BETA * BETA * BETA * BETA *
						BETA * BETA * BETA * BETA *
						BETA * BETA * BETA * BETA * B[i * (NJ) + j];
				}
    }
#endif
}

int main(int argc, char* argv[])
{
    num_cpu_cores = atoi(argv[1]);
    throttling_mod = atoi(argv[2]);
    throttling_alloc = atoi(argv[3]);

    NI = atoi(argv[4]);
    NJ = atoi(argv[5]);

#if COMPUTATION == 0
    computation = 0;
#elif COMPUTATION == 1
    computation = 1;
#elif COMPUTATION == 2
    computation = 2;
#elif COMPUTATION == 4
    computation = 4;
#elif COMPUTATION == 8
    computation = 8;
#elif COMPUTATION == 16
    computation = 16;
#endif

#if WORK_DIM == 1
    DIM_LOCAL_WORK_GROUP_X = atoi(argv[6]);
    DIM_LOCAL_WORK_GROUP_Y = 0;
    DIM_LOCAL_WORK_GROUP_Z = 0;
	gpu_percent = atoi(argv[7]);
    work_dim = 1;
#endif

    CreateProfiler();
    StartProfiling();

    A = (DATA_TYPE*)memalign(4096, NI*NJ*sizeof(DATA_TYPE));
    B = (DATA_TYPE*)memalign(4096, NI*NJ*sizeof(DATA_TYPE));
    C = (DATA_TYPE*)memalign(4096, NI*NJ*sizeof(DATA_TYPE));

    init(A, B, C);
    read_cl_file();
    cl_initialization();
    cl_mem_init(A, B, C);
    cl_load_prog();

    cl_launch_kernel();

    printf("Validation processing\n");
    DATA_TYPE* C_validation = (DATA_TYPE*)memalign(4096, NI*NJ*sizeof(DATA_TYPE));
    init(A, B, C_validation);
    mat2dadd(A, B, C_validation);
    compareResults(C, C_validation);

    cl_clean_up();

    free(A);
    free(B);
    free(C);
    free(C_validation);

    ReleaseProfiler();

    return 0;
}
