// MatAdd: Synthetic workloads (Matrix Addition) based on PolyBench/GPU 1.0 suite.
//
// Author: Younghyun Cho <younghyun@csap.snu.ac.kr>

#include <cstdlib>
#include <iostream>

#include <CL/cl.h>

#include <thread>
#include <vector>
#include <atomic>
#include <malloc.h>
#include <math.h>
#include <omp.h>

#include "profiler_interface.h"
#define MAX_SOURCE_SIZE (0x100000)

//define the error threshold for the results "not matching"
#define PERCENT_DIFF_ERROR_THRESHOLD 0.05

/* Declared constant values for ALPHA and BETA (same as values in PolyBench 2.0) */
//#define ALPHA 32412
//#define BETA 2123
#define ALPHA 3
#define BETA 2

/* Can switch DATA_TYPE between float and double */
#ifdef FLOAT
typedef float DATA_TYPE;
#else
#ifdef DOUBLE
typedef double DATA_TYPE;
#else
#ifdef INT
typedef int DATA_TYPE;
#endif
#endif
#endif

/* Workload description */
// D[i][j][k] = A[i][j][k] + B[C[k]]

/* Problem size */
int NI = 0;
int NJ = 0;
int NK = 0;

/* Amount of computational operations */
int computation = 1;

/* Thread block dimensions */
int DIM_LOCAL_WORK_GROUP_X = 16;
int DIM_LOCAL_WORK_GROUP_Y = 4;
int DIM_LOCAL_WORK_GROUP_Z = 2;
int work_dim = 2;

/* Data buffers */
DATA_TYPE* A;
DATA_TYPE* B;
int* C;
DATA_TYPE* D;

cl_platform_id platform_id;
cl_device_id device_id;   
cl_uint num_devices;
cl_uint num_platforms;
cl_int errcode;
cl_context clGPUContext;
cl_kernel clKernel;
cl_command_queue clCommandQue;
cl_program clProgram;
cl_mem a_mem_obj;
cl_mem b_mem_obj;
cl_mem c_mem_obj;
cl_mem d_mem_obj;
FILE *fp;
char *source_str;
size_t source_size;

#define SMALL_FLOAT_VAL 0.00000001f

float absVal(float a)
{
	if (a < 0) {
		return (a * -1);
	} else { 
		return a;
	}
}

float percentDiff(double val1, double val2)
{
	if ((absVal(val1) < 0.01) && (absVal(val2) < 0.01)) {
		return 0.0f;
    } else {
        return 100.0f * (absVal(absVal(val1 - val2) / absVal(val1 + SMALL_FLOAT_VAL)));
	}
} 

void compareResults(DATA_TYPE* D, DATA_TYPE* D_validation)
{
    int i, j, k, fail;
    fail = 0;

    // Compare C1 and C2
    for (i=0; i < NI; i++) {
        for (j=0; j < NJ; j++) {
            for (k=0; k < NK; k++) {
                if (percentDiff(D[i*(NJ*NK) + j*(NK) + k],
                                D_validation[i*(NJ*NK) + j*(NK) + k])
                        > PERCENT_DIFF_ERROR_THRESHOLD) {
                    fail++;
                }
            }
        }
    }

	// Print results
	printf("Non-Matching CPU-GPU Outputs Beyond Error Threshold of %4.2f Percent: %d\n", PERCENT_DIFF_ERROR_THRESHOLD, fail);
}

void read_cl_file()
{
	// Load the kernel source code into the array source_str
	fp = fopen("kernel.cl", "r");
	if (!fp) {
		fprintf(stderr, "Failed to load kernel.\n");
		exit(1);
	}
	source_str = (char*)malloc(MAX_SOURCE_SIZE);
	source_size = fread( source_str, 1, MAX_SOURCE_SIZE, fp);
	fclose( fp );
}

void init(DATA_TYPE* A, DATA_TYPE* B, int* C, DATA_TYPE* D)
{
    #pragma omp parallel for
    for (int i = 0; i < NI; i++) {
        for (int j = 0; j < NJ; j++) {
            for (int k = 0; k < NK; k++) {
                A[i * (NJ*NK) + j * NK + k] = (i * (NJ*NK) + j * NK + k)%10;
            }
        }
    }

    #pragma omp parallel for
    for (int i = 0; i < NI; i++) {
        for (int j = 0; j < NJ; j++) {
            for (int k = 0; k < NK; k++) {
                B[i * (NJ*NK) + j * NK + k] = (k * (NJ*NK) + j * NK + i)%10;
            }
        }
    }

    #pragma omp parallel for
    for (int k = 0; k < NK; k++) {
        C[k] = k*10;
    }

    #pragma omp parallel for
    for (int i = 0; i < NI; i++) {
        for (int j = 0; j < NJ; j++) {
            for (int k = 0; k < NK; k++) {
                D[i * (NJ*NK) + j * NK + k] = 0;
            }
        }
    }
}

void cl_initialization()
{	
    char str_temp[1024];

    // Get platform and device information
    errcode = clGetPlatformIDs(1, &platform_id, &num_platforms);
    if(errcode == CL_SUCCESS) printf("number of platforms is %d\n",num_platforms);
    else printf("Error getting platform IDs\n");

    errcode = clGetPlatformInfo(platform_id,CL_PLATFORM_NAME, sizeof(str_temp), str_temp,NULL);
    if(errcode == CL_SUCCESS) printf("platform name is %s\n",str_temp);
    else printf("Error getting platform name\n");

    errcode = clGetPlatformInfo(platform_id, CL_PLATFORM_VERSION, sizeof(str_temp), str_temp,NULL);
    if(errcode == CL_SUCCESS) printf("platform version is %s\n",str_temp);
    else printf("Error getting platform version\n");

    errcode = clGetDeviceIDs( platform_id, CL_DEVICE_TYPE_GPU, 1, &device_id, &num_devices);
    if(errcode == CL_SUCCESS) printf("number of devices is %d\n", num_devices);
    else printf("Error getting device IDs\n");

    errcode = clGetDeviceInfo(device_id,CL_DEVICE_NAME, sizeof(str_temp), str_temp,NULL);
    if(errcode == CL_SUCCESS) printf("device name is %s\n",str_temp);
    else printf("Error getting device name\n");

    // Create an OpenCL context
    clGPUContext = clCreateContext( NULL, 1, &device_id, NULL, NULL, &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating context\n");

    //Create a command-queue
    cl_queue_properties prop[] = { 0 };
    clCommandQue = clCreateCommandQueueWithProperties(clGPUContext, device_id, prop, &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating command queue\n");
}

void cl_mem_init(DATA_TYPE* A, DATA_TYPE* B, int* C, DATA_TYPE* D)
{
    a_mem_obj = clCreateBuffer(clGPUContext,
                               CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
                               sizeof(DATA_TYPE) * NI * NJ * NK,
                               A,
                               &errcode);
    b_mem_obj = clCreateBuffer(clGPUContext,
                               CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
                               sizeof(DATA_TYPE) * NI * NJ * NK,
                               B,
                               &errcode);
    c_mem_obj = clCreateBuffer(clGPUContext,
                               CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
                               sizeof(int) * NK,
                               C,
                               &errcode);
    d_mem_obj = clCreateBuffer(clGPUContext,
                               CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR,
                               sizeof(DATA_TYPE) * NI * NJ * NK,
                               D,
                               &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating buffers\n");
}

void cl_load_prog()
{
    // Create a program from the kernel source
    clProgram = clCreateProgramWithSource(clGPUContext, 1, (const char **)&source_str, (const size_t *)&source_size, &errcode);

    if(errcode != CL_SUCCESS) printf("Error in creating program\n");

    // Build the program
    char build_option[100] = { 0, };
#if defined(FLOAT)
    sprintf(build_option, "-I. -DFLOAT -DWORK_DIM=%d -DCOMPUTATION=%d", work_dim, computation);
#elif defined(DOUBLE)
    sprintf(build_option, "-I. -DDOUBLE -DWORK_DIM=%d -DCOMPUTATION=%d", work_dim, computation);
#elif defined(INT)
    sprintf(build_option, "-I. -DINT -DWORK_DIM=%d -DCOMPUTATION=%d", work_dim, computation);
#endif

    errcode = clBuildProgram(clProgram, 1, &device_id, build_option, NULL, NULL);
    if(errcode != CL_SUCCESS) printf("Error in building program\n");

    // Create the OpenCL kernel
    clKernel = clCreateKernel(clProgram, "mat3dadd", &errcode);
    if(errcode != CL_SUCCESS) printf("Error in creating kernel\n");
    clFinish(clCommandQue);
}

void cl_launch_kernel()
{
    int ni=NI;
    int nj=NJ;
    int nk=NK;

    DATA_TYPE alpha = ALPHA;
    DATA_TYPE beta = BETA;

#if WORK_DIM == 3
    size_t localWorkSize[3], globalWorkSize[3];
    localWorkSize[0] = DIM_LOCAL_WORK_GROUP_X;
    localWorkSize[1] = DIM_LOCAL_WORK_GROUP_Y;
    localWorkSize[2] = DIM_LOCAL_WORK_GROUP_Z;
    globalWorkSize[0] = (size_t)ceil(((float)NI) / ((float)DIM_LOCAL_WORK_GROUP_X)) * DIM_LOCAL_WORK_GROUP_X;
    globalWorkSize[1] = (size_t)ceil(((float)NJ) / ((float)DIM_LOCAL_WORK_GROUP_Y)) * DIM_LOCAL_WORK_GROUP_Y;
    globalWorkSize[2] = (size_t)ceil(((float)NK) / ((float)DIM_LOCAL_WORK_GROUP_Z)) * DIM_LOCAL_WORK_GROUP_Z;
#elif WORK_DIM == 2
    size_t localWorkSize[2], globalWorkSize[2];
    localWorkSize[0] = DIM_LOCAL_WORK_GROUP_X;
    localWorkSize[1] = DIM_LOCAL_WORK_GROUP_Y;
    globalWorkSize[0] = (size_t)ceil(((float)NI) / ((float)DIM_LOCAL_WORK_GROUP_X)) * DIM_LOCAL_WORK_GROUP_X;
    globalWorkSize[1] = (size_t)ceil(((float)NJ) / ((float)DIM_LOCAL_WORK_GROUP_Y)) * DIM_LOCAL_WORK_GROUP_Y;
#elif WORK_DIM == 1
    size_t localWorkSize[1], globalWorkSize[1];
    localWorkSize[0] = DIM_LOCAL_WORK_GROUP_X;
    globalWorkSize[0] = (size_t)ceil(((float)NI) / ((float)DIM_LOCAL_WORK_GROUP_X)) * DIM_LOCAL_WORK_GROUP_X;
#endif

    // Set the arguments of the kernel
    errcode =  clSetKernelArg(clKernel, 0, sizeof(cl_mem), (void*)&a_mem_obj);
    errcode |= clSetKernelArg(clKernel, 1, sizeof(cl_mem), (void*)&b_mem_obj);
    errcode |= clSetKernelArg(clKernel, 2, sizeof(cl_mem), (void*)&c_mem_obj);
    errcode |= clSetKernelArg(clKernel, 3, sizeof(cl_mem), (void*)&d_mem_obj);
    errcode |= clSetKernelArg(clKernel, 4, sizeof(DATA_TYPE), (void*)&alpha);
    errcode |= clSetKernelArg(clKernel, 5, sizeof(DATA_TYPE), (void*)&beta);
    errcode |= clSetKernelArg(clKernel, 6, sizeof(int), (void*)&ni);
    errcode |= clSetKernelArg(clKernel, 7, sizeof(int), (void*)&nj);
    errcode |= clSetKernelArg(clKernel, 8, sizeof(int), (void*)&nk);
    if(errcode != CL_SUCCESS) printf("Error in setting arguments\n");

    StartProfiling();
    errcode = clEnqueueNDRangeKernel(clCommandQue, clKernel, work_dim, NULL,
                                     globalWorkSize, localWorkSize, 0, NULL, NULL);
    if (errcode != CL_SUCCESS) printf("Error in launching kernel\n");
    clFlush(clCommandQue);
    clFinish(clCommandQue);
    StopProfiling("mat3dadd", "profiling.txt");
}

void cl_clean_up()
{
    // Clean up
    errcode = clFlush(clCommandQue);
    errcode = clFinish(clCommandQue);
    errcode = clReleaseKernel(clKernel);
    errcode = clReleaseProgram(clProgram);
    errcode = clReleaseMemObject(a_mem_obj);
    errcode = clReleaseMemObject(b_mem_obj);
    errcode = clReleaseMemObject(c_mem_obj);
    errcode = clReleaseCommandQueue(clCommandQue);
    errcode = clReleaseContext(clGPUContext);
    if(errcode != CL_SUCCESS) printf("Error in cleanup\n");
}

void mat3dadd(DATA_TYPE* A, DATA_TYPE* B, int* C, DATA_TYPE* D)
{
#if COMPUTATION == 0
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
            for (k = 0; k < NK; ++k) {
                D[i * (NJ*NK) + j * (NK) + k] = A[i * (NJ*NK) + j * (NK) + k] +
                                                B[C[k]];
            }
        }
    }
#elif COMPUTATION == 1
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
            for (k = 0; k < NK; ++k) {
                D[i * (NJ*NK) + j * (NK) + k] = ALPHA * A[i * (NJ*NK) + j * (NK) + k] +
                                                BETA * B[C[k]];
            }
        }
    }
#elif COMPUTATION == 2
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
            for (k = 0; k < NK; ++k) {
                D[i * (NJ*NK) + j * (NK) + k] = ALPHA * ALPHA * A[i * (NJ*NK) + j * (NK) + k] +
                                                BETA * BETA * B[C[k]];
            }
        }
    }
#elif COMPUTATION == 4
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
            for (k = 0; k < NK; ++k) {
                D[i * (NJ*NK) + j * (NK) + k] = ALPHA * ALPHA * ALPHA * ALPHA * A[i * (NJ*NK) + j * (NK) + k] +
                                                BETA * BETA * BETA * BETA * B[C[k]];
            }
        }
    }
#elif COMPUTATION == 8
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
            for (k = 0; k < NK; ++k) {
                D[i * (NJ*NK) + j * (NK) + k] = ALPHA * ALPHA * ALPHA * ALPHA *
                                                ALPHA * ALPHA * ALPHA * ALPHA * A[i * (NJ*NK) + j * (NK) + k] +
                                                BETA * BETA * BETA * BETA *
                                                BETA * BETA * BETA * BETA * B[C[k]];
            }
        }
    }
#elif COMPUTATION == 16
    int i,j,k;

    #pragma omp parallel for private(i,j,k)
    for (i = 0; i < NI; i++) {
        for (j = 0; j < NJ; j++) {
            for (k = 0; k < NK; ++k) {
                D[i * (NJ*NK) + j * (NK) + k] = ALPHA * ALPHA * ALPHA * ALPHA *
                                                ALPHA * ALPHA * ALPHA * ALPHA *
                                                ALPHA * ALPHA * ALPHA * ALPHA *
                                                ALPHA * ALPHA * ALPHA * ALPHA * A[i * (NJ*NK) + j * (NK) + k] +
                                                BETA * BETA * BETA * BETA *
                                                BETA * BETA * BETA * BETA *
                                                BETA * BETA * BETA * BETA *
                                                BETA * BETA * BETA * BETA * B[C[k]];
            }
        }
    }
#endif
}

int main(int argc, char* argv[])
{
    NI = atoi(argv[1]);
    NJ = atoi(argv[2]);
    NK = atoi(argv[3]);

#if COMPUTATION == 0
    computation = 0;
#elif COMPUTATION == 1
    computation = 1;
#elif COMPUTATION == 2
    computation = 2;
#elif COMPUTATION == 4
    computation = 4;
#elif COMPUTATION == 8
    computation = 8;
#elif COMPUTATION == 16
    computation = 16;
#endif

#if WORK_DIM == 3
    DIM_LOCAL_WORK_GROUP_X = atoi(argv[4]);
    DIM_LOCAL_WORK_GROUP_Y = atoi(argv[5]);
    DIM_LOCAL_WORK_GROUP_Z = atoi(argv[6]);
    work_dim = 3;
#elif WORK_DIM == 2
    DIM_LOCAL_WORK_GROUP_X = atoi(argv[4]);
    DIM_LOCAL_WORK_GROUP_Y = atoi(argv[5]);
    DIM_LOCAL_WORK_GROUP_Z = 0;
    work_dim = 2;
#elif WORK_DIM == 1
    DIM_LOCAL_WORK_GROUP_X = atoi(argv[4]);
    DIM_LOCAL_WORK_GROUP_Y = 0;
    DIM_LOCAL_WORK_GROUP_Z = 0;
    work_dim = 1;
#endif

    CreateProfiler();

    A = (DATA_TYPE*)memalign(4096, NI*NJ*NK*sizeof(DATA_TYPE));
    B = (DATA_TYPE*)memalign(4096, NI*NJ*NK*sizeof(DATA_TYPE));
    C = (int*)memalign(4096, NK*sizeof(int));
    D = (DATA_TYPE*)memalign(4096, NI*NJ*NK*sizeof(DATA_TYPE));

    init(A, B, C, D);
    read_cl_file();
    cl_initialization();
    cl_mem_init(A, B, C, D);
    cl_load_prog();

    cl_launch_kernel();

    //printf("Validation processing\n");
    //DATA_TYPE* D_validation = (DATA_TYPE*)memalign(4096, NI*NJ*NK*sizeof(DATA_TYPE));
    //init(A, B, C, D_validation);
    //mat3dadd(A, B, C, D_validation);
    //compareResults(D, D_validation);
    //free(D_validation);

    cl_clean_up();

    free(A);
    free(B);
    free(C);
    free(D);

    ReleaseProfiler();

    return 0;
}
